import json
import logging
from logging import Logger
from typing import List, Optional, Dict, AsyncGenerator

from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.responses.fhir_update_response import FhirUpdateResponse

from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.utilities.fhir_helpers.get_fhir_client import (
    get_fhir_client,
)


async def send_json_bundle_to_fhir_async(
    *,
    id_: Optional[str],
    json_data_list: List[str],
    server_url: str,
    validation_server_url: Optional[str],
    resource: str,
    logger: Logger,
    auth_server_url: Optional[str],
    auth_client_id: Optional[str],
    auth_client_secret: Optional[str],
    auth_login_token: Optional[str],
    auth_scopes: Optional[List[str]],
    auth_access_token: Optional[str],
    auth_well_known_url: Optional[str],
    log_level: Optional[str],
    retry_count: Optional[int] = None,
    exclude_status_codes_from_retry: Optional[List[int]] = None,
    additional_request_headers: Optional[Dict[str, str]] = None,
    batch_size: Optional[int] = None,
    smart_merge: Optional[bool] = None,
) -> AsyncGenerator[FhirMergeResponse, None]:
    """
    Send a JSON bundle to FHIR server

    """
    assert id_, f"{json_data_list!r}"
    fhir_client: FhirClient = get_fhir_client(
        logger=logger,
        server_url=server_url,
        auth_server_url=auth_server_url,
        auth_client_id=auth_client_id,
        auth_client_secret=auth_client_secret,
        auth_login_token=auth_login_token,
        auth_access_token=auth_access_token,
        auth_scopes=auth_scopes,
        log_level=log_level,
        auth_well_known_url=auth_well_known_url,
    )
    if smart_merge is False:
        fhir_client.smart_merge(False)

    fhir_client = fhir_client.resource(resource)
    if validation_server_url:
        fhir_client = fhir_client.validation_server_url(validation_server_url)

    if retry_count is not None:
        fhir_client = fhir_client.retry_count(retry_count)
    if additional_request_headers is not None:
        logger.info(
            f"Additional Request Headers to be sent - {additional_request_headers}"
        )
        fhir_client = fhir_client.additional_request_headers(additional_request_headers)

    if exclude_status_codes_from_retry:
        fhir_client = fhir_client.exclude_status_codes_from_retry(
            exclude_status_codes_from_retry
        )

    try:
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("----------- Sending data to FHIR -------")
            logger.debug(json_data_list)
            logger.debug("----------- End sending data to FHIR -------")
        response: FhirMergeResponse
        async for response in fhir_client.merge_async(
            id_=id_, json_data_list=json_data_list, batch_size=batch_size
        ):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("----------- Response from FHIR -------")
                logger.debug(f"{response.status} {response.url}")
                logger.debug(f"{response.error}")
                logger.debug(f"{response.data}")
                logger.debug("----------- End response from FHIR -------")
            yield response
    except AssertionError as e:
        logger.exception(
            Exception(
                f"Assertion: FHIR send failed: {str(e)} for resource: {json_data_list}"
            )
        )
        yield FhirMergeResponse(
            request_id="",
            status=400,
            error=str(e),
            json_data=json.dumps(json_data_list),
            access_token=auth_access_token,
            url=server_url,
            responses=[],
        )


async def send_fhir_delete_async(
    obj_id: str,
    server_url: str,
    resource: str,
    logger: Logger,
    auth_server_url: Optional[str],
    auth_client_id: Optional[str],
    auth_client_secret: Optional[str],
    auth_login_token: Optional[str],
    auth_scopes: Optional[List[str]],
    auth_access_token: Optional[str],
    auth_well_known_url: Optional[str],
    log_level: Optional[str],
    additional_request_headers: Optional[Dict[str, str]] = None,
) -> AsyncGenerator[FhirDeleteResponse, None]:
    fhir_client: FhirClient = get_fhir_client(
        logger=logger,
        server_url=server_url,
        auth_server_url=auth_server_url,
        auth_client_id=auth_client_id,
        auth_client_secret=auth_client_secret,
        auth_login_token=auth_login_token,
        auth_access_token=auth_access_token,
        auth_scopes=auth_scopes,
        log_level=log_level,
        auth_well_known_url=auth_well_known_url,
    )

    fhir_client = fhir_client.resource(resource)
    fhir_client.resource(resource).id_(obj_id)
    if additional_request_headers is not None:
        logger.info(
            f"Additional Request Headers to be sent - {additional_request_headers}"
        )
        fhir_client = fhir_client.additional_request_headers(additional_request_headers)
    try:
        response: FhirDeleteResponse = await fhir_client.delete_async()
        yield response
    except AssertionError as e:
        logger.exception(
            Exception(f"Assertion: FHIR send failed: {str(e)} for {resource}: {obj_id}")
        )
        yield FhirDeleteResponse(
            request_id="",
            status=400,
            error=str(e),
            access_token=auth_access_token,
            url=server_url,
            responses=json.dumps([]),
            resource_type=resource,
        )


async def update_json_bundle_to_fhir_async(
    *,
    obj_id: str,
    json_data: str,
    server_url: str,
    operation: FhirSenderOperation | str,
    validation_server_url: Optional[str],
    resource: str,
    logger: Logger,
    auth_server_url: Optional[str],
    auth_client_id: Optional[str],
    auth_client_secret: Optional[str],
    auth_login_token: Optional[str],
    auth_scopes: Optional[List[str]],
    auth_access_token: Optional[str],
    auth_well_known_url: Optional[str],
    log_level: Optional[str],
    retry_count: Optional[int] = None,
    exclude_status_codes_from_retry: Optional[List[int]] = None,
    additional_request_headers: Optional[Dict[str, str]] = None,
) -> AsyncGenerator[FhirUpdateResponse, None]:
    assert obj_id, f"{json_data!r}"
    fhir_client: FhirClient = get_fhir_client(
        logger=logger,
        server_url=server_url,
        auth_server_url=auth_server_url,
        auth_client_id=auth_client_id,
        auth_client_secret=auth_client_secret,
        auth_login_token=auth_login_token,
        auth_access_token=auth_access_token,
        auth_scopes=auth_scopes,
        log_level=log_level,
        auth_well_known_url=auth_well_known_url,
    )

    fhir_client = fhir_client.resource(resource)
    fhir_client.resource(resource).id_(obj_id)
    if validation_server_url:
        fhir_client = fhir_client.validation_server_url(validation_server_url)

    if retry_count is not None:
        fhir_client = fhir_client.retry_count(retry_count)
    if additional_request_headers is not None:
        logger.info(
            f"Additional Request Headers to be sent - {additional_request_headers}"
        )
        fhir_client = fhir_client.additional_request_headers(additional_request_headers)

    if exclude_status_codes_from_retry:
        fhir_client = fhir_client.exclude_status_codes_from_retry(
            exclude_status_codes_from_retry
        )

    try:
        logger.debug("----------- Sending data to FHIR -------")
        logger.debug(json_data)
        logger.debug("----------- End sending data to FHIR -------")
        response: FhirUpdateResponse
        if FhirSenderOperation.operation_equals(
            operation, FhirSenderOperation.FHIR_OPERATION_PATCH
        ):
            response = await fhir_client.send_patch_request_async(data=json_data)
            yield response
        else:
            response = await fhir_client.update_async(json_data=json_data)
            yield response
    except AssertionError as e:
        logger.exception(
            Exception(
                f"Assertion: FHIR update failed: {str(e)} for resource: {json_data}"
            )
        )
        yield FhirUpdateResponse(
            request_id="",
            error=str(e),
            access_token=auth_access_token,
            status=400,
            url=server_url,
            responses=json.dumps([]),
            resource_type=resource,
        )



import json
from dataclasses import dataclass, asdict
from typing import Optional

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)


@dataclass
class OAuth2Credentails:
    client_id: str
    client_secret: str
    grant_type: Optional[str] = "client_credentials"
    audience: Optional[str] = None


class OAuth2ClientCredentialsFlow:
    def __init__(
        self,
        *,
        auth_url: str,
        auth_credentails: OAuth2Credentails,
        progress_logger: Optional[ProgressLogger],
    ) -> None:
        """
        Encapsulate the logic to connect to OAuth2 servers using client credentials flow

        :param auth_url: OAuth2 token url
        :param auth_credentails: OAuth2 credentails like client id, secrets, etc
        :param progress_logger: Progress logger
        """
        assert auth_url
        self.auth_url: str = auth_url
        self.auth_credentails = auth_credentails
        self.progress_logger: Optional[ProgressLogger] = progress_logger

    def get_token(self) -> Optional[str]:
        http_request = HelixHttpRequest(
            url=self.auth_url,
            payload=asdict(
                self.auth_credentails,
                dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
            ),
            request_type=RequestType.POST,
        )

        response = http_request.get_result()
        if self.progress_logger:
            self.progress_logger.write_to_log(
                f"Received from {self.auth_url}: {json.dumps(response.result)}"
            )
        token = response.result.get("access_token")

        return token



from pyspark.sql.session import SparkSession

from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_list_catalog_table_names,
)


class SparkTestHelper:
    @staticmethod
    def clear_tables(spark_session: SparkSession) -> None:
        """
        :param spark_session:
        """

        # spark_session.sql("SET -v").show(n=200, truncate=False)

        table_names = spark_list_catalog_table_names(spark_session)

        for table_name in table_names:
            print(f"clear_tables() dropping table/view: {table_name}")
            spark_session.sql(f"DROP TABLE IF EXISTS default.{table_name}")
            spark_session.sql(f"DROP VIEW IF EXISTS default.{table_name}")
            spark_session.sql(f"DROP VIEW IF EXISTS {table_name}")

        spark_session.catalog.clearCache()



from pyspark.sql.session import SparkSession

from spark_pipeline_framework.transformers.framework_jdbc_reader.v1.framework_jdbc_reader import (
    FrameworkJdbcReader,
)


def test_can_read_via_jdbc(spark_session: SparkSession) -> None:
    """
    Because testing against a database is a pain and most of the core transformation logic is in the base module,
    we're only testing that the options and format are correctly being exposed.
    """
    # Arrange
    view = "my_view"
    name = "read_my_data"
    jdbc_url = "jdbc:mysql:user@password:host/db:port"
    query = "(select * from questionnaire) questionnaire_alias"
    driver = "org.driver.FakeDriver"

    # Act
    reader = FrameworkJdbcReader(
        jdbc_url=jdbc_url, query=query, driver=driver, view=view, name=name
    )

    # Assert
    assert reader.getJdbcUrl() == jdbc_url
    assert reader.getDriver() == driver
    assert reader.getFormat() == "jdbc"
    assert reader.getQuery() == query
    assert reader.getView() == view



import json
import math
from typing import Any, Dict, Optional

from pyspark import RDD, StorageLevel
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import Row, StructType, StructField, StringType, IntegerType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.transformers.http_data_sender.v1.http_data_sender_processor import (
    HttpDataSenderProcessor,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.oauth2_helpers.v1.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class HttpDataSender(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        source_view: str,
        view: Optional[str] = None,
        url: Optional[str] = None,
        auth_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        parse_response_as_json: Optional[bool] = True,
        content_type: str = "application/x-www-form-urlencoded",
        post_as_json_formatted_string: Optional[bool] = None,
        batch_count: Optional[int] = None,
        batch_size: Optional[int] = None,
        cache_storage_level: Optional[StorageLevel] = None,
    ):
        """
        Sends data to http server (usually REST API)


        :param source_view: read the payload (body) from this view
        :param view: (Optional) view to put the responses in
        :param url: url to call
        :param auth_url: (Optional) url to use to authenticate with client credentials
        :param client_id: (Optional) client id to use to authenticate with client credentials
        :param client_secret: (Optional) client secret to use to authenticate with client credentials
        :param parse_response_as_json: (Optional) whether to parse response as json or not (default = True)
        :param content_type: content_type to use when posting
        :param batch_count: (Optional) number of batches to create
        :param batch_size: (Optional) max number of items in a batch
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=source_view)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.url: Param[Optional[str]] = Param(self, "url", "")
        self._setDefault(url=url)

        self.auth_url: Param[Optional[str]] = Param(self, "auth_url", "")
        self._setDefault(auth_url=auth_url)

        self.client_id: Param[Optional[str]] = Param(self, "client_id", "")
        self._setDefault(client_id=client_id)

        self.client_secret: Param[Optional[str]] = Param(self, "client_secret", "")
        self._setDefault(client_secret=client_secret)

        self.content_type: Param[str] = Param(self, "content_type", "")
        self._setDefault(content_type=content_type)

        self.batch_count: Param[Optional[int]] = Param(self, "batch_count", "")
        self._setDefault(batch_count=batch_count)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.parse_response_as_json: Param[Optional[bool]] = Param(
            self, "parse_response_as_json", ""
        )
        self._setDefault(parse_response_as_json=parse_response_as_json)

        self.post_as_json_formatted_string: Param[Optional[bool]] = Param(
            self, "post_as_json_formatted_string", ""
        )
        self._setDefault(post_as_json_formatted_string=post_as_json_formatted_string)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=cache_storage_level)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        name: Optional[str] = self.getName()
        auth_url: Optional[str] = self.getOrDefault(self.auth_url)
        client_id: Optional[str] = self.getOrDefault(self.client_id)
        client_secret: Optional[str] = self.getOrDefault(self.client_secret)
        progress_logger = self.getProgressLogger()
        source_view: str = self.getOrDefault(self.source_view)
        view: Optional[str] = self.getOrDefault(self.view)
        url: Optional[str] = self.getOrDefault(self.url)
        parse_response_as_json: Optional[bool] = self.getOrDefault(
            self.parse_response_as_json
        )
        content_type: str = self.getOrDefault(self.content_type)
        batch_count: Optional[int] = self.getOrDefault(self.batch_count)
        batch_size: Optional[int] = self.getOrDefault(self.batch_size)
        post_as_json_formatted_string: Optional[bool] = self.getOrDefault(
            self.post_as_json_formatted_string
        )
        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )

        df = df.sparkSession.table(source_view)

        if spark_is_data_frame_empty(df=df):
            return df

        headers: Dict[str, Any] = {}

        if client_id and auth_url and client_secret:
            # first call auth to get a token
            oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
                OAuth2ClientCredentialsFlow(
                    auth_url=auth_url,
                    client_id=client_id,
                    client_secret=client_secret,
                    progress_logger=progress_logger,
                )
            )

            access_token: Optional[str] = oauth2_client_credentials_flow.get_token()

            if progress_logger:
                progress_logger.write_to_log(
                    f"Received token from {auth_url}: {access_token}"
                )

            if access_token:
                headers = {"Authorization": f"Bearer {access_token}"}

        if progress_logger:
            progress_logger.write_to_log(
                f"Using headers: {json.dumps(headers, default=str)}"
            )

        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            desired_partitions: int
            if batch_count:
                desired_partitions = batch_count
            else:
                row_count: int = df.count()
                desired_partitions = (
                    math.ceil(row_count / batch_size)
                    if batch_size and batch_size > 0
                    else row_count
                )
            self.logger.info(
                f"Total Batches: {desired_partitions} for total rows: {row_count}"
            )

            # ---- Now process all the results ----
            rdd: RDD[Row] = df.repartition(
                desired_partitions
            ).rdd.mapPartitionsWithIndex(
                lambda partition_index, rows: HttpDataSenderProcessor.send_partition_to_server(
                    partition_index=partition_index,
                    rows=rows,
                    url=url,
                    content_type=content_type,
                    headers=headers,
                    post_as_json_formatted_string=post_as_json_formatted_string,
                    parse_response_as_json=parse_response_as_json,
                )
            )

            rdd = (
                rdd.cache()
                if cache_storage_level is None
                else rdd.persist(storageLevel=cache_storage_level)
            )

            response_schema: StructType = StructType(
                [
                    StructField("url", StringType(), True),
                    StructField("status", IntegerType(), True),
                    StructField("result", StringType(), True),
                    StructField("headers", StringType(), True),
                    StructField("payload", StringType(), True),
                    StructField("request_type", StringType(), True),
                ]
            )

            try:
                result_df: DataFrame = (
                    (rdd.toDF())  # let Spark auto-discover the schema
                    if parse_response_as_json
                    else (rdd.toDF(schema=response_schema))
                )

                result_df = result_df.where(col("url").isNotNull())
                if view:
                    result_df.createOrReplaceTempView(view)
            except Exception as e:
                raise Exception(
                    f"Unable to read data frame: "
                    + f"{','.join([json.dumps(row.asDict(recursive=True)) for row in rdd.collect()])}"
                ) from e

            return result_df



from os import environ
from pathlib import Path
from typing import Any, Dict, Optional, Union, List

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import to_json, struct, col
from pyspark.errors import AnalysisException, PythonException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_processor import (
    ElasticSearchProcessor,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_result import (
    ElasticSearchResult,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_sender_parameters import (
    ElasticSearchSenderParameters,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class ElasticSearchSender(FrameworkTransformer):
    @capture_parameters
    def __init__(
        self,
        index: str,
        file_path: Optional[Union[Path, str]] = None,
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        batch_size: int = 100,
        limit: int = -1,
        multi_line: bool = False,
        operation: str = "index",
        output_path: Optional[Union[Path, str]] = None,
        run_synchronously: Optional[bool] = None,
        log_level: Optional[str] = None,
        timeout: int = 60,
        max_chunk_size: int = 100,
        process_chunks_in_parallel: Optional[bool] = True,
        maximum_concurrent_tasks: int = 100,
    ):
        """
        Sends a folder or a view to an ElasticSearch server
        Either file_path or view must be set

        :param index: ElasticSearch index to update
        :param file_path: (Optional) read data from this folder to send
        :param view: (Optional) read data from this view
        :param batch_size: how many documents to process at one time
        :param multi_line: when reading data from folder, whether to expect multiline json files
        :param output_path: writes bulk output to this folder before sending to ES
        :param run_synchronously: if True, will send all data to ES in one go, otherwise will send in batches
        :param log_level: log level
        :param timeout: timeout in seconds
        :param max_chunk_size: the size of the chunks
        :param process_chunks_in_parallel: whether to run all the chunks in a partition in parallel (default is sequential)
        :param maximum_concurrent_tasks: maximum number of tasks to run concurrently (default is 100)
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert file_path or view

        assert progress_logger

        self.log_level: Param[str] = Param(self, "log_level", "")
        self._setDefault(log_level=log_level)

        self.logger = get_logger(__name__, level=log_level or "INFO")

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[Optional[Union[Path, str]]] = Param(self, "file_path", "")
        self._setDefault(file_path=file_path)

        self.index: Param[str] = Param(self, "index", "")
        self._setDefault(index=index)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=limit)

        self.batch_size: Param[int] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.multi_line: Param[bool] = Param(self, "multi_line", "")
        self._setDefault(multi_line=multi_line)

        self.operation: Param[str] = Param(self, "operation", "")
        self._setDefault(operation=operation)

        self.output_path: Param[Optional[Union[Path, str]]] = Param(
            self, "output_path", ""
        )
        self._setDefault(output_path=output_path)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.timeout: Param[int] = Param(self, "timeout", "")
        self._setDefault(timeout=timeout)

        self.max_chunk_size: Param[int] = Param(self, "max_chunk_size", "")
        self._setDefault(max_chunk_size=max_chunk_size)

        self.process_chunks_in_parallel: Param[Optional[bool]] = Param(
            self, "process_chunks_in_parallel", ""
        )
        self._setDefault(process_chunks_in_parallel=process_chunks_in_parallel)

        self.maximum_concurrent_tasks: Param[int] = Param(
            self, "maximum_concurrent_tasks", ""
        )
        self._setDefault(maximum_concurrent_tasks=maximum_concurrent_tasks)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        path: Optional[Union[Path, str]] = self.getFilePath()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        index: str = self.getIndex()
        batch_size: int = self.getBatchSize()
        limit: Optional[int] = self.getLimit()
        operation: str = self.getOperation()
        parameters: Optional[Dict[str, Any]] = self.getParameters()
        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)
        log_level: Optional[str] = self.getOrDefault(self.log_level) or environ.get(
            "LOGLEVEL"
        )
        timeout: int = self.getOrDefault(self.timeout)

        max_chunk_size: int = self.getOrDefault(self.max_chunk_size)
        process_chunks_in_parallel: Optional[bool] = self.getOrDefault(
            self.process_chunks_in_parallel
        )
        maximum_concurrent_tasks: int = self.getOrDefault(self.maximum_concurrent_tasks)

        doc_id_prefix: Optional[str] = None
        if parameters is not None:
            doc_id_prefix = parameters.get("doc_id_prefix", None)

        # limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_elasticsearch_sender", progress_logger=progress_logger
        ):
            # read all the files at path into a dataframe
            path_to_files: str = str(path)
            try:
                if view:
                    json_df: DataFrame = df.sparkSession.table(view)
                    if "value" not in json_df.columns:
                        json_df = json_df.select(
                            to_json(struct(col("*"))).alias("value")
                        )
                else:
                    json_df = df.sparkSession.read.text(
                        path_to_files, pathGlobFilter="*.json", recursiveFileLookup=True
                    )
                if limit is not None and limit > 0:
                    json_df = json_df.limit(limit)
                # json_df.show(truncate=False, n=100000)
            except AnalysisException as e:
                if str(e).startswith("Path does not exist:"):
                    if progress_logger:
                        progress_logger.write_to_log(f"Folder {path_to_files} is empty")
                    return df
                raise

            row_count: int = json_df.count()
            if not spark_is_data_frame_empty(df=json_df):
                self.logger.info(
                    f"----- Sending {index} (rows={row_count}) to ElasticSearch server -----"
                )
                total_partitions: int = json_df.rdd.getNumPartitions()
                self.logger.info(f"----- Total Batches: {total_partitions}  -----")

                sender_parameters: ElasticSearchSenderParameters = (
                    ElasticSearchSenderParameters(
                        index=index,
                        operation=operation,
                        total_partitions=total_partitions,
                        doc_id_prefix=doc_id_prefix,
                        name=name,
                        log_level=log_level,
                        timeout=timeout,
                        pandas_udf_parameters=AsyncPandasUdfParameters(
                            log_level=log_level,
                            max_chunk_size=max_chunk_size,
                            process_chunks_in_parallel=process_chunks_in_parallel,
                            maximum_concurrent_tasks=maximum_concurrent_tasks,
                        ),
                    )
                )
                if run_synchronously:
                    rows_to_send: List[Dict[str, Any]] = [
                        r.asDict(recursive=True) for r in json_df.collect()
                    ]
                    result_rows: List[ElasticSearchResult] = (
                        await AsyncHelper.collect_items(
                            ElasticSearchProcessor.send_partition_to_server_async(
                                partition_index=0,
                                rows=rows_to_send,
                                parameters=sender_parameters,
                            )
                        )
                    )
                    result_dicts: List[Dict[str, Any]] = [
                        r.to_dict_flatten_payload()
                        for r in result_rows
                        if r is not None
                    ]
                    result_df = (
                        df.sparkSession.createDataFrame(  # type:ignore[type-var]
                            result_dicts, schema=ElasticSearchResult.get_schema()
                        )
                    )
                else:
                    # use mapInPandas
                    # https://spark.apache.org/docs/latest/api/python/user_guide/sql/arrow_pandas.html#map
                    # https://docs.databricks.com/en/pandas/pandas-function-apis.html#map
                    # Source Code: https://github.com/apache/spark/blob/master/python/pyspark/sql/pandas/map_ops.py#L37
                    result_df = json_df.mapInPandas(
                        ElasticSearchProcessor.get_process_partition_function(
                            parameters=sender_parameters,
                            batch_size=batch_size,
                        ),
                        schema=ElasticSearchResult.get_schema(),
                    )

                # select just the columns needed to minimize extra data having to be pulled back to the driver
                success_df: DataFrame = result_df.where(col("failed") == 0).select(
                    "url", "success"
                )
                failed_df: DataFrame = result_df.where(col("failed") > 0).select(
                    "url", "failed", "payload", "error"
                )

                try:
                    self.logger.info("---- Success ----")
                    success_df.show(truncate=False, n=1000)
                    self.logger.info("---- Failed ----")
                    failed_df.show(truncate=False, n=1000)
                    self.logger.info("---- End Reply from server ----")
                except PythonException as e:
                    # noinspection PyUnresolvedReferences
                    if hasattr(e, "desc") and "pyarrow.lib.ArrowTypeError" in e.desc:
                        raise FriendlySparkException(
                            exception=e,
                            message="Exception converting data to Arrow format."
                            + f" This is usually because the return data did not match the specified schema.",
                            stage_name=name,
                        )
                    else:
                        raise
                except Exception as e:
                    raise FriendlySparkException(exception=e, stage_name=name)

        self.logger.info(
            f"----- Finished sending {index} (rows={row_count}) to ElasticSearch server  -----"
        )
        return result_df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIndex(self) -> str:
        return self.getOrDefault(self.index)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> int:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiline(self) -> int:
        return self.getOrDefault(self.multi_line)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or f"{self.getView()} - {self.getIndex()}"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOperation(self) -> str:
        return self.getOrDefault(self.operation)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputPath(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.output_path)



from typing import Any, Dict, Optional

from spark_pipeline_framework.logger.log_level import LogLevel

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.pretty_print import (
    get_pretty_data_frame,
    get_data_frame_as_csv,
)


class FrameworkLogSqlResultTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        sql: Optional[str] = None,
        view: Optional[str] = None,
        limit: Optional[int] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        log_event: Optional[bool] = None,
        log_event_level: Optional[LogLevel] = None,
        log_limit: Optional[int] = None,
    ) -> None:
        """
        Logs the result after running sql or the contents of the view


        :param sql: SQL to run
        :param view: store results of running SQL into this view
        :param log_event: whether to log an event at TRACE level with progress_logger with the contents of the output view
        :param log_event_level: the level to use for logging the event
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.sql: Param[Optional[str]] = Param(self, "sql", "")
        self._setDefault(sql=sql)
        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        assert sql or view

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=limit)

        self.log_event: Param[Optional[bool]] = Param(self, "log_event", "")
        self._setDefault(log_event=None)

        self.log_event_level: Param[Optional[LogLevel]] = Param(
            self, "log_event_level", ""
        )
        self._setDefault(log_event_level=None)

        self.log_limit: Param[Optional[int]] = Param(self, "log_limit", "")
        self._setDefault(log_limit=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        name = self.getName()
        sql: Optional[str] = self.getSql()
        view: Optional[str] = self.getOrDefault(self.view)
        limit = self.getLimit()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        log_event: Optional[bool] = self.getOrDefault(self.log_event)
        log_event_level: Optional[LogLevel] = self.getOrDefault(self.log_event_level)
        log_limit: Optional[int] = self.getOrDefault(self.log_limit)

        df2: DataFrame
        if sql:
            df2 = df.sparkSession.sql(sql)
        elif view:
            df2 = df.sparkSession.table(view)
        else:
            assert False, "Neither sql nor view was specified"

        log_limit = 10 if log_limit is None else log_limit

        if log_limit and log_limit > 0:
            message = (
                "\n"
                + (name or view or "")
                + f" (LIMIT {log_limit})"
                + "\n"
                + get_pretty_data_frame(df2, log_limit, name=sql or view)
            )
            self.logger.info(message)
            if progress_logger:
                progress_logger.write_to_log(message)
                if log_event or log_event_level:
                    if log_event_level:
                        progress_logger.log_event(
                            event_name=name or view or self.__class__.__name__,
                            event_text=message,
                            log_level=log_event_level,
                        )
                    else:
                        progress_logger.log_event(
                            event_name=name or view or self.__class__.__name__,
                            event_text=message,
                        )
                progress_logger.log_artifact(
                    key=f"{name or view}.csv",
                    contents=get_data_frame_as_csv(df=df2, limit=log_limit),
                )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> Optional[str]:
        return self.getOrDefault(self.sql)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from mockserver_client.mock_requests_loader import load_mock_source_api_json_responses
from mockserver_client.mockserver_client import MockServerFriendlyClient
from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.http_data_sender.v1.http_data_sender import (
    HttpDataSender,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_http_data_sender_parse_response_as_json(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_name = "test_http_data_sender"

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # noinspection PyTypeChecker
    spark_session.createDataFrame(  # type:ignore[type-var]
        [
            {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43919",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
            {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43910",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
        ],
    ).createOrReplaceTempView("my_view")

    server_url = f"http://mock-server:1080/{test_name}"
    mock_client = MockServerFriendlyClient("http://mock-server:1080")
    mock_client.clear(test_name)
    mock_client.reset()

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("token_service"),
        mock_client=mock_client,
        url_prefix=test_name,
    )

    # Act
    with ProgressLogger() as progress_logger:
        HttpDataSender(
            progress_logger=progress_logger,
            source_view="my_view",
            view="output_view",
            url=f"{server_url}",
            parse_response_as_json=True,
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("output_view")
    result_df.printSchema()
    result_df.show(truncate=False)

    # Collect the DataFrame and convert each Row to a dictionary
    result_dicts = [row.asDict() for row in result_df.collect()]

    # convert result, headers and payload to json
    for result_dict in result_dicts:
        result_dict["headers"] = json.loads(result_dict["headers"])
        result_dict["payload"] = json.loads(result_dict["payload"])

    assert [
        r
        for r in result_dicts
        if r["payload"]["member_id"] == "e3a3f665-eae5-4046-a241-efdfe5c43919"
    ] == [
        {
            "headers": {"Content-Type": "application/x-www-form-urlencoded"},
            "payload": {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43919",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
            "request_type": "RequestType.POST",
            "result": {
                "access_token": "fake access_token",
                "expires_in": "54000",
                "token_type": "bearer",
            },
            "status": 200,
            "url": "http://mock-server:1080/test_http_data_sender",
        }
    ]
    assert [
        r
        for r in result_dicts
        if r["payload"]["member_id"] == "e3a3f665-eae5-4046-a241-efdfe5c43910"
    ] == [
        {
            "headers": {"Content-Type": "application/x-www-form-urlencoded"},
            "payload": {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43910",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
            "request_type": "RequestType.POST",
            "result": {
                "access_token": "fake access_token2",
                "expires_in": "54000",
                "token_type": "bearer",
            },
            "status": 200,
            "url": "http://mock-server:1080/test_http_data_sender",
        }
    ]



import json
from typing import Dict, List, Optional, Any, AsyncGenerator, Protocol, Type

import aiohttp
from aiohttp import ClientResponseError

from spark_pipeline_framework.utilities.aws.config import get_ssm_config
import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendor_response_key_error import (
    VendorResponseKeyError,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.melissa_standardizing_vendor_api_response import (
    MelissaStandardizingVendorApiResponse,
)

logger = structlog.get_logger(__file__)


class CustomApiCallFunction(Protocol):
    async def __call__(
        self,
        raw_addresses: List[RawAddress],
    ) -> Dict[str, Any]:
        """
        This function is called with a batch of raw addresses and should return a response from the vendor

        :param raw_addresses: raw addresses as a list
        :return: output values as a list of dictionaries
        """
        ...


class MelissaStandardizingVendor(
    StandardizingVendor[MelissaStandardizingVendorApiResponse]
):
    def batch_request_max_size(self) -> int:
        return 100

    @classmethod
    def get_api_response_class(cls) -> Type[MelissaStandardizingVendorApiResponse]:
        return MelissaStandardizingVendorApiResponse

    def __init__(
        self,
        license_key: str = "",
        custom_api_call: Optional[CustomApiCallFunction] = None,
        version: str = "1",
        api_calls_limit: Optional[int] = None,
        response_key_error_threshold: int = 2,
    ) -> None:
        """
        This class is responsible for standardizing addresses using Melissa API

        :param license_key: License key for Melissa API
        :param custom_api_call: Custom API call to Melissa API
        :param version: Version of the response
        :param api_calls_limit: Maximum number of calls to Melissa API allowed
        :param response_key_error_threshold: Number of times Melissa is allowed to send bad response
                                            until we cancel rest of requests
        """
        super().__init__(version)
        assert (
            license_key or custom_api_call
        ), "Either license_key or custom_api_call must be provided"
        self._license_key: str = license_key
        self._custom_api_call_async: Optional[CustomApiCallFunction] = custom_api_call
        self._error_counter: int = 0
        self._api_calls_limit: Optional[int] = api_calls_limit
        self._api_calls_counter: int = 0
        self._response_key_error_threshold: int = response_key_error_threshold

    async def standardize_async(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse[MelissaStandardizingVendorApiResponse]]:
        """
        returns the vendor specific response from the vendor
        """
        vendor_specific_addresses: List[MelissaStandardizingVendorApiResponse] = []
        async for vendor_specific_addresses1 in self._call_std_addr_api_async(
            raw_addresses=raw_addresses
        ):
            vendor_specific_addresses.extend(vendor_specific_addresses1)

        vendor_responses: List[
            VendorResponse[MelissaStandardizingVendorApiResponse]
        ] = self._to_vendor_response(
            vendor_response=vendor_specific_addresses,
            raw_addresses=raw_addresses,
            vendor_name=self.get_vendor_name(),
            response_version=self.get_version(),
        )
        return vendor_responses

    def vendor_specific_to_std(
        self,
        *,
        vendor_specific_addresses: List[
            VendorResponse[MelissaStandardizingVendorApiResponse]
        ],
        raw_addresses: List[RawAddress],
    ) -> List[StandardizedAddress]:
        """
        Each vendor class knows how to convert its response to StdAddress
        Note: to interpret the standardize address. reference (https://www.melissa.com/quickstart-guides/global-address)
        AddressLine1-2-3 are not consistent from address to address so we infer line1 from FormattedAddress and keep
        Line2 empty for now
        """

        std_addresses = [
            a.api_call_response.to_standardized_address(
                address_id=a.api_call_response.RecordID
            )
            for a in vendor_specific_addresses
        ]
        return std_addresses

    async def _call_std_addr_api_async(
        self, raw_addresses: List[RawAddress]
    ) -> AsyncGenerator[List[MelissaStandardizingVendorApiResponse], None]:
        """
        Please check https://www.melissa.com/quickstart-guides/global-address for more info
        Please make sure "License Key" is available https://www.melissa.com/user/user_account.aspx
        """

        if len(raw_addresses) == 0:
            logger.info("No addresses to standardize")
            yield []
            return

        self._api_calls_counter += 1

        if self._api_calls_limit and self._api_calls_counter > self._api_calls_limit:
            logger.error(
                f"Melissa API calls limit reached. Limit: {self._api_calls_limit}"
            )
            yield [
                MelissaStandardizingVendorApiResponse.from_standardized_address(
                    StandardizedAddress.from_raw_address(
                        raw_address=raw_address, vendor_name=self.get_vendor_name()
                    )
                )
                for raw_address in raw_addresses
            ]
            return

        try:
            api_server_response: Dict[str, Any] = (
                await self._custom_api_call_async(raw_addresses)
                if self._custom_api_call_async
                else await self._api_call_async(raw_addresses)
            )

            # adding vendor, so we can parse it correctly after reading response from cache in the future
            vendor_specific_addresses: List[Dict[str, str]] | None = (
                api_server_response.get("Records")
            )
            if vendor_specific_addresses is None:
                raise VendorResponseKeyError
            else:
                # we may not get records from all the addresses we sent
                yield [
                    MelissaStandardizingVendorApiResponse.from_dict(v)
                    for v in vendor_specific_addresses
                ]
        except (ClientResponseError, VendorResponseKeyError) as e:
            logger.exception(
                f"{self.get_vendor_name()} response does not have a Records key. Are we out of credit? Error: {e}"
            )
            raise e

    async def _api_call_async(self, raw_addresses: List[RawAddress]) -> Dict[str, Any]:
        addresses_to_lookup: List[Dict[str, str]] = [
            {
                "RecordID": a.get_id() or "",
                "AddressLine1": a.to_str(),
                "Country": a.to_dict()["country"],
            }
            for a in raw_addresses
        ]

        license_key = self._get_request_credentials()["license_key"]
        url = r"https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress"
        json_batch_dict = {
            "TransmissionReference": "GlobalAddressBatch",
            "CustomerID": license_key,
            "Options": "",
            "Records": addresses_to_lookup,
        }
        payload = json.dumps(json_batch_dict).encode("utf-8")
        headers = {"Accept": "application/json", "Content-Type": "application/json"}

        try:
            async with aiohttp.ClientSession() as session:
                logger.debug(f"Making request to {url} {json_batch_dict}")
                async with session.post(url, headers=headers, data=payload) as response:
                    logger.debug(f"Response status: {response.status}")
                    response.raise_for_status()
                    api_response: Dict[str, List[Dict[str, str]]] = (
                        await response.json()
                    )
                    logger.debug(f"Response: {api_response}")
        except Exception as e:
            logger.exception(
                f"Error connecting to {self.get_vendor_name()}", http_error=repr(e)
            )
            raise e

        return api_response

    def _get_request_credentials(self) -> Dict[str, str]:
        if self._license_key:
            license_key = self._license_key
        else:
            base_path = "/prod/helix/external/melissa/"
            c = get_ssm_config(base_path)
            license_key = c[f"{base_path}license_key_credit"]
        return {"license_key": license_key}

    @classmethod
    def get_vendor_name(cls) -> str:
        return "melissa"

    @staticmethod
    def get_record_id(record: Dict[str, str]) -> str:
        return record["RecordID"]

    def _to_vendor_response(
        self,
        vendor_response: List[MelissaStandardizingVendorApiResponse],
        raw_addresses: List[RawAddress],
        vendor_name: str,
        response_version: str,
    ) -> List[VendorResponse[MelissaStandardizingVendorApiResponse]]:
        # create the map
        id_response_map = {a.get_id(): a for a in raw_addresses}
        # find and assign
        return [
            VendorResponse(
                api_call_response=r,
                related_raw_address=id_response_map[r.RecordID],
                vendor_name=vendor_name,
                response_version=response_version,
            )
            for r in vendor_response
        ]



from typing import Optional


class LoopIdMixin:
    def __init__(self) -> None:
        self.loop_id: Optional[str] = None

    def set_loop_id(self, loop_id: Optional[str]) -> None:
        """
        Set when running inside a FrameworkLoopTransformer

        :param loop_id: loop id
        """
        self.loop_id = loop_id



from os import environ
from typing import Dict, Any, Optional, Type, Union, List, Callable

from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent

from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkExceptionHandlerTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        exception_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = None,
        raise_on_exception: Optional[Union[bool, Callable[[DataFrame], bool]]] = True,
        handled_exceptions: Optional[List[Type[BaseException]]] = None,
        retry_count: Optional[int] = 0,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Executes a sequence of stages (transformers) and, in case of an exception, executes a separate
        sequence of exception-handling stages.

        :param: name: Name of the transformer.
        :param: stages: The primary sequence of transformers to execute.
        :param: exception_stages: Stages to execute if an error occurs.
        :param: raise_on_exception: Determines whether to raise exceptions when errors occur.
        :param: handled_exceptions: List of exception types to catch. Defaults to [BaseException]
        :param: retry_count: Number of times to retry a failed stage before running exception stages or raising exception.
        :param: parameters: Additional parameters.
        :param: progress_logger: Logger instance for tracking execution.

        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.raise_on_exception: Optional[Union[bool, Callable[[DataFrame], bool]]] = (
            raise_on_exception
        )

        self.retry_count: int = retry_count or 0

        self.handled_exceptions: List[Type[BaseException]] = handled_exceptions or [
            BaseException
        ]
        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages
        self.exception_stages: Union[
            List[Transformer], Callable[[], List[Transformer]]
        ] = (exception_stages or [])

        self.loop_id: Optional[str] = None

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        """
        Executes the transformation pipeline asynchronously with retry logic.

        - Runs `stages` normally with retry logic.
        - If an exception occurs and retries are exhausted, executes `exception_stages` if provided.
        - Optionally raises exceptions based on `raise_on_exception`.
        """
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        stage_name = ""
        raise_on_exception = (
            self.raise_on_exception
            if not callable(self.raise_on_exception)
            else self.raise_on_exception(df)
        )

        async def run_single_stage(
            df: DataFrame,
            stage: Transformer,
            stage_name: str,
            telemetry_span_creator: TelemetrySpanCreator,
        ) -> DataFrame:
            async with telemetry_span_creator.create_telemetry_span_async(
                name=stage_name,
                attributes={},
                telemetry_parent=self.telemetry_parent,
            ) as telemetry_span:

                if isinstance(stage, LoopIdMixin):
                    stage.set_loop_id(self.loop_id)
                if isinstance(stage, TelemetryParentMixin):
                    stage.set_telemetry_parent(
                        telemetry_parent=telemetry_span.create_child_telemetry_parent()
                    )

                if hasattr(stage, "transform_async"):
                    return await stage.transform_async(df)  # type: ignore
                else:
                    return stage.transform(df)

        async def run_stages_with_retry(
            df: DataFrame,
            stages: Union[List[Transformer], Callable[[], List[Transformer]]],
            retry_count: int,
            stage_type: str = "stage",
        ) -> DataFrame:
            """Run stages with retry logic."""
            stages_list = stages if not callable(stages) else stages()
            nonlocal stage_name

            telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
                telemetry_parent=self.telemetry_parent
                or TelemetryParent.get_null_parent()
            ).create_telemetry_span_creator(log_level=environ.get("LOGLEVEL"))

            for stage in stages_list:
                if hasattr(stage, "getName"):
                    stage_name = stage.getName()
                    if stage_name is not None:
                        stage_name = f"{stage_name} ({stage.__class__.__name__})"
                    stage_name = stage.__class__.__name__
                attempt = 0

                while attempt <= retry_count:
                    try:
                        if progress_logger and attempt > 0:
                            progress_logger.write_to_log(
                                self.getName()
                                or "FrameworkExceptionHandlerTransformer",
                                f"Retrying {stage_type} {stage_name} (attempt {attempt + 1}/{retry_count + 1})",
                            )

                        df = await run_single_stage(
                            df, stage, stage_name, telemetry_span_creator
                        )
                        break  # Stage succeeded, exit retry loop

                    except Exception as e:
                        if len(e.args) >= 1:
                            e.args = (f"In Stage ({stage_name})", *e.args)

                        if progress_logger:
                            progress_logger.write_to_log(
                                self.getName()
                                or "FrameworkExceptionHandlerTransformer",
                                f"{stage_name} failed on attempt {attempt + 1}/{retry_count + 1}. "
                                f"Exception: {type(e).__name__}: {str(e)}",
                            )

                        attempt += 1
                        if attempt > retry_count:  # Exhausted retries
                            raise e

            return df

        def is_handled_exception(exception: Exception) -> bool:
            """Check if exception is one of the handled exception types"""
            return any(
                isinstance(exception, exception_class)
                for exception_class in self.handled_exceptions
            )

        try:
            df = await run_stages_with_retry(df, self.stages, self.retry_count, "stage")
        except Exception as e:
            is_exception_handled = is_handled_exception(e)

            if progress_logger:
                progress_logger.write_to_log(
                    self.getName() or "FrameworkExceptionHandlerTransformer",
                    f"Exception occurred in stage: {stage_name} after {self.retry_count + 1} attempts. "
                    f"Exception type: {type(e).__name__}. "
                    f"Is handled exception: {is_exception_handled}.",
                )

            # Store the failed stage name before it potentially gets updated in exception stages
            failed_stage_name = stage_name

            # Run exception stages if the exception is handled and exception stages are defined
            if is_exception_handled and self.exception_stages:
                try:
                    df = await run_stages_with_retry(
                        df, self.exception_stages, self.retry_count, "exception stage"
                    )
                except Exception as err:
                    if progress_logger:
                        progress_logger.write_to_log(
                            self.getName() or "FrameworkExceptionHandlerTransformer",
                            f"Exception occurred in exception handling stage: {stage_name}. "
                            f"Exception: {type(err).__name__}: {str(err)}",
                        )
                    raise err

            # Decide whether to re-raise the original exception
            if raise_on_exception or not is_exception_handled:
                e.args = (f"In Stage ({failed_stage_name})", *e.args)
                raise e
            else:
                if progress_logger:
                    progress_logger.write_to_log(
                        self.getName() or "FrameworkExceptionHandlerTransformer",
                        f"Suppressing exception from stage: {failed_stage_name}. "
                        f"Exception was handled and raise_on_exception=False",
                    )

        return df

    def as_dict(self) -> Dict[str, Any]:

        return {
            **(super().as_dict()),
            "raise_on_exception": self.raise_on_exception,
            "retry_count": self.retry_count,
            "stages": (
                [s.as_dict() for s in self.stages]  # type: ignore
                if not callable(self.stages)
                else str(self.stages)
            ),
            "exception_stages": (
                [s.as_dict() for s in self.exception_stages]  # type: ignore
                if self.exception_stages and not callable(self.exception_stages)
                else str(self.exception_stages)
            ),
        }



import json
from os import path, makedirs, environ
from pathlib import Path
from shutil import rmtree
from urllib.parse import urljoin

import pytest
from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.utilities.fhir_server_helpers import FhirServerHelpers
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import IntegerType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender import FhirSender
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

import requests


@pytest.mark.parametrize("run_synchronously", [True, False])
async def test_fhir_sender_merge(
    spark_session: SparkSession, run_synchronously: bool
) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    await FhirServerHelpers.clean_fhir_server_async(resource_type="Patient")

    test_files_dir: Path = data_dir.joinpath("test_files/patients")
    response_files_dir: Path = temp_folder.joinpath("patients-response")
    parameters = {"flow_name": "Test Pipeline V2"}

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url

        # first delete any existing resources
        fhir_client = await fhir_server_test_context.create_fhir_client_async()

        fhir_client = fhir_client.url(fhir_server_url).resource("Patient")
        delete_response: FhirDeleteResponse = await fhir_client.id_(
            "00100000000"
        ).delete_async()
        assert delete_response.status == 204
        delete_response = await fhir_client.id_("00200000000").delete_async()
        assert delete_response.status == 204

        token_url = fhir_server_test_context.get_token_url()
        assert token_url

        authorization_header = fhir_server_test_context.get_authorization_header()

        environ["LOGLEVEL"] = "DEBUG"
        # Act
        with ProgressLogger() as progress_logger:
            await FhirSender(
                resource="Patient",
                server_url=fhir_server_url,
                file_path=test_files_dir,
                response_path=response_files_dir,
                progress_logger=progress_logger,
                batch_size=1,
                run_synchronously=run_synchronously,
                additional_request_headers={"SampleHeader": "SampleValue"},
                parameters=parameters,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_well_known_url=auth_well_known_url,
                view="result_view",
                error_view="error_view",
            ).transform_async(df)

        # Assert
        error_df = spark_session.read.table("error_view")
        result_df = spark_session.read.table("result_view")
        # Assert
        assert error_df.count() == 0
        assert result_df.count() == 2

        response = requests.get(
            urljoin(fhir_server_url, "Patient/00100000000"),
            headers=authorization_header,
        )
        assert response.ok, response.text
        json_text: str = response.text
        obj = json.loads(json_text)
        assert obj["birthDate"] == "2017-01-01"

        response = requests.get(
            urljoin(fhir_server_url, "Patient/00200000000"),
            headers=authorization_header,
        )
        assert response.ok, response.text
        json_text = response.text
        obj = json.loads(json_text)
        assert obj["birthDate"] == "1984-01-01"


@pytest.mark.parametrize("run_synchronously", [True])
async def test_fhir_sender_merge_for_custom_parameters(
    spark_session: SparkSession, run_synchronously: bool
) -> None:
    print("")
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:
        test_files_dir: Path = data_dir.joinpath("test_files/eob_with_added_field")
        response_files_dir: Path = temp_folder.joinpath("eob-with-added-field-response")

        df: DataFrame = create_empty_dataframe(spark_session=spark_session)

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url
        token_url = fhir_server_test_context.get_token_url()
        assert token_url
        authorization_header = fhir_server_test_context.get_authorization_header()

        fhir_client: FhirClient = (
            await fhir_server_test_context.create_fhir_client_async()
        )

        fhir_client = fhir_client.url(fhir_server_url)
        fhir_client.resource("ExplanationOfBenefit")

        await fhir_client.id_("H111-12345").delete_async()

        await fhir_client.id_("H222-12345").delete_async()

        await fhir_client.id_("H333-12345").delete_async()

        environ["LOGLEVEL"] = "DEBUG"
        # Act
        with ProgressLogger() as progress_logger:
            await FhirSender(
                resource="ExplanationOfBenefit",
                server_url=fhir_server_url,
                file_path=test_files_dir,
                response_path=response_files_dir,
                progress_logger=progress_logger,
                run_synchronously=run_synchronously,
                sort_by_column_name_and_type=("source_file_line_num", IntegerType()),
                drop_fields_from_json=["source_file_line_num"],
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_well_known_url=auth_well_known_url,
            ).transform_async(df)

            # for first EOB
            response = requests.get(
                urljoin(fhir_server_url, "ExplanationOfBenefit/H111-12345"),
                headers=authorization_header,
            )
            assert response.ok, response.text
            json_text = response.text
            obj = json.loads(json_text)
            # verify that latest value has been appended for benefit amount
            assert obj["item"][0]["adjudication"][0]["amount"]["value"] == 17
            assert obj.get("source_file_line_num") is None

            # for second EOB
            response = requests.get(
                urljoin(fhir_server_url, "ExplanationOfBenefit/H222-12345"),
                headers=authorization_header,
            )
            assert response.ok, response.text
            json_text = response.text
            print(json_text)
            obj = json.loads(json_text)

            # verify that latest value has been appended for benefit amount
            assert obj["item"][0]["adjudication"][0]["amount"]["value"] == 100
            assert obj.get("source_file_line_num") is None

            # for third EOB
            response = requests.get(
                urljoin(fhir_server_url, "ExplanationOfBenefit/H333-12345"),
                headers=authorization_header,
            )
            assert response.ok, response.text
            json_text = response.text
            obj = json.loads(json_text)
            # verify that latest value has been appended for benefit amount
            assert obj["item"][0]["adjudication"][0]["amount"]["value"] == 17
            assert obj.get("source_file_line_num") is None



import asyncio
import json
from datetime import datetime
from json import JSONDecodeError
from logging import Logger
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
    cast,
    NamedTuple,
    Coroutine,
)

from furl import furl

from helix_fhir_client_sdk.exceptions.fhir_sender_exception import FhirSenderException
from compressedfhir.fhir.fhir_resource_list import FhirResourceList
from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.filters.sort_field import SortField
from helix_fhir_client_sdk.responses.fhir_get_response import FhirGetResponse
from compressedfhir.utilities.fhir_json_encoder import FhirJSONEncoder
from pyspark.sql.types import (
    Row,
)
from helix_fhir_client_sdk.function_types import RefreshTokenFunction

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_response_writer import (
    FhirGetResponseWriter,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_parser_exception import (
    FhirParserException,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_receiver_exception import (
    FhirReceiverException,
)
from spark_pipeline_framework.utilities.fhir_helpers.get_fhir_client import (
    get_fhir_client,
)
from spark_pipeline_framework.utilities.flattener.flattener import flatten


class GetBatchResult(NamedTuple):
    resources: List[str]
    errors: List[str]


class FhirReceiverHelpers:
    @staticmethod
    # function that is called for each partition
    def send_partition_request_to_server(
        *,
        partition_index: int,
        rows: Iterable[Row],
        batch_size: Optional[int],
        has_token_col: bool,
        server_url: Optional[str],
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        auth_access_token: Optional[str],
        resource_type: str,
        error_view: Optional[str],
        url_column: Optional[str],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction] = None,
    ) -> List[Row]:
        """
        This function processes a partition calling fhir server for each row in the partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        :param partition_index: partition index
        :param rows: rows to process
        :param batch_size: batch size
        :param has_token_col: has token column
        :param server_url: server url
        :param log_level: log level
        :param action: action to perform
        :param action_payload: action payload
        :param additional_parameters: additional parameters
        :param filter_by_resource: filter by resource
        :param filter_parameter: filter parameter
        :param sort_fields: sort fields
        :param auth_server_url: auth server url
        :param auth_client_id: auth client id
        :param auth_client_secret: auth client secret
        :param auth_login_token: auth login token
        :param auth_scopes: auth scopes
        :param include_only_properties: include only properties
        :param separate_bundle_resources: separate bundle resources
        :param expand_fhir_bundle: expand fhir bundle
        :param accept_type: accept type
        :param content_type: content type
        :param additional_request_headers: Additional request headers
        :param accept_encoding: accept encoding
        :param slug_column: slug column
        :param retry_count: retry count
        :param exclude_status_codes_from_retry: exclude status codes from retry
        :param limit: limit
        :param auth_access_token: auth access token
        :param resource_type: resource type
        :param error_view: error view
        :param url_column: url column
        :param use_data_streaming: use data streaming
        :param graph_json: graph json
        :param refresh_token_function: function to refresh token
        :return: rows
        """
        register()
        resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
            (
                {
                    "resource_id": r["id"],
                    "access_token": r["token"],
                    url_column: r[url_column],
                    slug_column: r[slug_column],
                    "resourceType": r["resourceType"],
                }
                if has_token_col and url_column and slug_column
                else (
                    {
                        "resource_id": r["id"],
                        "access_token": r["token"],
                    }
                    if has_token_col
                    else {"resource_id": r["id"], "access_token": auth_access_token}
                )
            )
            for r in rows
        ]
        result = FhirReceiverHelpers.process_with_token(
            partition_index=partition_index,
            resource_id_with_token_list=resource_id_with_token_list,
            batch_size=batch_size,
            has_token_col=has_token_col,
            server_url=server_url,
            log_level=log_level,
            action=action,
            action_payload=action_payload,
            additional_parameters=additional_parameters,
            filter_by_resource=filter_by_resource,
            filter_parameter=filter_parameter,
            sort_fields=sort_fields,
            auth_server_url=auth_server_url,
            auth_client_id=auth_client_id,
            auth_client_secret=auth_client_secret,
            auth_login_token=auth_login_token,
            auth_scopes=auth_scopes,
            include_only_properties=include_only_properties,
            separate_bundle_resources=separate_bundle_resources,
            expand_fhir_bundle=expand_fhir_bundle,
            accept_type=accept_type,
            content_type=content_type,
            additional_request_headers=additional_request_headers,
            accept_encoding=accept_encoding,
            slug_column=slug_column,
            retry_count=retry_count,
            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
            limit=limit,
            auth_access_token=auth_access_token,
            resource_type=resource_type,
            error_view=error_view,
            url_column=url_column,
            use_data_streaming=use_data_streaming,
            graph_json=graph_json,
            refresh_token_function=refresh_token_function,
        )
        return result

    @staticmethod
    def process_with_token(
        *,
        partition_index: int,
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        batch_size: Optional[int],
        has_token_col: bool,
        server_url: Optional[str],
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        auth_access_token: Optional[str],
        resource_type: str,
        error_view: Optional[str],
        url_column: Optional[str],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> List[Row]:
        try:
            first_id: Optional[str] = resource_id_with_token_list[0]["resource_id"]
        except IndexError:
            first_id = None

        try:
            last_id: Optional[str] = resource_id_with_token_list[-1]["resource_id"]
        except IndexError:
            last_id = None

        sent: int = len(resource_id_with_token_list)

        if sent == 0:
            return [
                FhirGetResponseWriter.create_row(
                    partition_index=partition_index,
                    sent=0,
                    received=0,
                    responses=[],
                    first=None,
                    last=None,
                    error_text=None,
                    url=server_url,
                    status_code=200,
                    request_id=None,
                    access_token=None,
                    extra_context_to_return=None,
                )
            ]

        # if batch and not has_token then send all ids at once as long as the access token is the same
        if batch_size and batch_size > 1 and not has_token_col:
            return FhirReceiverHelpers.process_batch(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                resource_id_with_token_list=resource_id_with_token_list,
                server_url=server_url,
                log_level=log_level,
                action=action,
                action_payload=action_payload,
                additional_parameters=additional_parameters,
                filter_by_resource=filter_by_resource,
                filter_parameter=filter_parameter,
                sort_fields=sort_fields,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                include_only_properties=include_only_properties,
                separate_bundle_resources=separate_bundle_resources,
                expand_fhir_bundle=expand_fhir_bundle,
                accept_type=accept_type,
                content_type=content_type,
                additional_request_headers=additional_request_headers,
                accept_encoding=accept_encoding,
                slug_column=slug_column,
                retry_count=retry_count,
                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                limit=limit,
                auth_access_token=auth_access_token,
                resource_type=resource_type,
                error_view=error_view,
                use_data_streaming=use_data_streaming,
                graph_json=graph_json,
                refresh_token_function=refresh_token_function,
            )
        else:  # otherwise send one by one
            return FhirReceiverHelpers.process_one_by_one(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                resource_id_with_token_list=resource_id_with_token_list,
                server_url=server_url,
                log_level=log_level,
                action=action,
                action_payload=action_payload,
                additional_parameters=additional_parameters,
                filter_by_resource=filter_by_resource,
                filter_parameter=filter_parameter,
                sort_fields=sort_fields,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                include_only_properties=include_only_properties,
                separate_bundle_resources=separate_bundle_resources,
                expand_fhir_bundle=expand_fhir_bundle,
                accept_type=accept_type,
                content_type=content_type,
                additional_request_headers=additional_request_headers,
                accept_encoding=accept_encoding,
                slug_column=slug_column,
                retry_count=retry_count,
                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                limit=limit,
                error_view=error_view,
                url_column=url_column,
                resource_name=resource_type,
                use_data_streaming=use_data_streaming,
                graph_json=graph_json,
                refresh_token_function=refresh_token_function,
            )

    @staticmethod
    def process_one_by_one(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        server_url: Optional[str],
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        error_view: Optional[str],
        url_column: Optional[str],
        resource_name: str,
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> List[Row]:
        coroutines: List[Coroutine[Any, Any, List[Row]]] = [
            FhirReceiverHelpers.process_single_row_async(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                server_url=server_url,
                log_level=log_level,
                action=action,
                action_payload=action_payload,
                additional_parameters=additional_parameters,
                filter_by_resource=filter_by_resource,
                filter_parameter=filter_parameter,
                sort_fields=sort_fields,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                include_only_properties=include_only_properties,
                separate_bundle_resources=separate_bundle_resources,
                expand_fhir_bundle=expand_fhir_bundle,
                accept_type=accept_type,
                content_type=content_type,
                additional_request_headers=additional_request_headers,
                accept_encoding=accept_encoding,
                slug_column=slug_column,
                retry_count=retry_count,
                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                limit=limit,
                error_view=error_view,
                url_column=url_column,
                resource1=resource1,
                resource_name=resource_name,
                use_data_streaming=use_data_streaming,
                graph_json=graph_json,
                refresh_token_function=refresh_token_function,
            )
            for resource1 in resource_id_with_token_list
        ]

        async def get_functions() -> List[List[Row]]:
            # noinspection PyTypeChecker
            return await asyncio.gather(*[asyncio.create_task(c) for c in coroutines])

        result: List[List[Row]] = AsyncHelper.run(get_functions())
        return flatten(result)

    @staticmethod
    async def process_single_row_async(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        server_url: Optional[str],
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        error_view: Optional[str],
        url_column: Optional[str],
        resource_name: str,
        resource1: Dict[str, Optional[str]],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> List[Row]:
        id_ = resource1["resource_id"]
        access_token = resource1["access_token"]
        url_ = resource1.get(url_column) if url_column else None
        service_slug = resource1.get(slug_column) if slug_column else None
        resource_type = resource1.get("resourceType")
        request_id: Optional[str] = None
        responses_from_fhir: List[str] = []
        extra_context_to_return: Optional[Dict[str, Any]] = None
        try:
            result1: FhirGetResponse = (
                await FhirReceiverHelpers.send_simple_fhir_request_async(
                    id_=id_,
                    token_=access_token,
                    server_url_=url_ or server_url,
                    service_slug=service_slug,
                    resource_type=resource_type or resource_name,
                    log_level=log_level,
                    server_url=server_url,
                    action=action,
                    action_payload=action_payload,
                    additional_parameters=additional_parameters,
                    filter_by_resource=filter_by_resource,
                    filter_parameter=filter_parameter,
                    sort_fields=sort_fields,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    include_only_properties=include_only_properties,
                    separate_bundle_resources=separate_bundle_resources,
                    expand_fhir_bundle=expand_fhir_bundle,
                    accept_type=accept_type,
                    content_type=content_type,
                    additional_request_headers=additional_request_headers,
                    accept_encoding=accept_encoding,
                    slug_column=slug_column,
                    retry_count=retry_count,
                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                    limit=limit,
                    use_data_streaming=use_data_streaming,
                    graph_json=graph_json,
                    refresh_token_function=refresh_token_function,
                )
            )
            resp_result: str = result1.get_response_text().replace("\n", "")
            try:
                responses_from_fhir = FhirReceiverHelpers.json_str_to_list_str(
                    resp_result
                )
            except JSONDecodeError as e2:
                if error_view:
                    result1.error = f"{(result1.error or '')}: {str(e2)}"
                else:
                    raise FhirParserException(
                        url=result1.url,
                        message="Parsing result as json failed",
                        json_data=result1.get_response_text(),
                        response_status_code=result1.status,
                        request_id=result1.request_id,
                    ) from e2

            error_text = result1.error
            status_code = result1.status
            request_url: Optional[str] = result1.url
            request_id = result1.request_id
            extra_context_to_return = result1.extra_context_to_return
            access_token = result1.access_token
        except FhirSenderException as e1:
            error_text = str(e1)
            status_code = e1.response_status_code or 0
            request_url = e1.url

        result = [
            FhirGetResponseWriter.create_row(
                partition_index=partition_index,
                sent=1,
                received=len(responses_from_fhir),
                responses=responses_from_fhir,
                first=first_id,
                last=last_id,
                error_text=error_text,
                url=request_url,
                status_code=status_code,
                request_id=request_id,
                extra_context_to_return=extra_context_to_return,
                access_token=access_token,
            )
        ]
        return result

    @staticmethod
    def process_batch(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        server_url: Optional[str],
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        auth_access_token: Optional[str],
        resource_type: str,
        error_view: Optional[str],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> List[Row]:
        result1 = AsyncHelper.run(
            FhirReceiverHelpers.send_simple_fhir_request_async(
                id_=[cast(str, r["resource_id"]) for r in resource_id_with_token_list],
                token_=auth_access_token,
                server_url=server_url,
                server_url_=server_url,
                resource_type=resource_type,
                log_level=log_level,
                action=action,
                action_payload=action_payload,
                additional_parameters=additional_parameters,
                filter_by_resource=filter_by_resource,
                filter_parameter=filter_parameter,
                sort_fields=sort_fields,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                include_only_properties=include_only_properties,
                separate_bundle_resources=separate_bundle_resources,
                expand_fhir_bundle=expand_fhir_bundle,
                accept_type=accept_type,
                content_type=content_type,
                additional_request_headers=additional_request_headers,
                accept_encoding=accept_encoding,
                slug_column=slug_column,
                retry_count=retry_count,
                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                limit=limit,
                use_data_streaming=use_data_streaming,
                graph_json=graph_json,
                refresh_token_function=refresh_token_function,
            )
        )
        resp_result: str = result1.get_response_text().replace("\n", "")
        responses_from_fhir = []
        try:
            responses_from_fhir = FhirReceiverHelpers.json_str_to_list_str(resp_result)
        except JSONDecodeError as e1:
            if error_view:
                result1.error = f"{(result1.error or '')}: {str(e1)}"
            else:
                raise FhirParserException(
                    url=result1.url,
                    message="Parsing result as json failed",
                    json_data=result1.get_response_text(),
                    response_status_code=result1.status,
                    request_id=result1.request_id,
                ) from e1

        error_text = result1.error
        status_code = result1.status
        request_id: Optional[str] = result1.request_id
        is_valid_response: bool = True if len(responses_from_fhir) > 0 else False
        result = [
            FhirGetResponseWriter.create_row(
                partition_index=partition_index,
                sent=1,
                received=len(responses_from_fhir) if is_valid_response else 0,
                responses=responses_from_fhir if is_valid_response else [],
                first=first_id,
                last=last_id,
                error_text=error_text,
                url=result1.url,
                status_code=status_code,
                request_id=request_id,
                access_token=None,
                extra_context_to_return=None,
            )
        ]
        return result

    @staticmethod
    async def send_simple_fhir_request_async(
        *,
        id_: Optional[Union[str, List[str]]],
        token_: Optional[str],
        server_url: Optional[str],
        server_url_: Optional[str],
        service_slug: Optional[str] = None,
        resource_type: str,
        log_level: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        include_only_properties: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        slug_column: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        limit: Optional[int],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> FhirGetResponse:
        url = server_url_ or server_url
        assert url
        return await FhirReceiverHelpers.send_fhir_request_async(
            logger=get_logger(__name__),
            action=action,
            action_payload=action_payload,
            additional_parameters=additional_parameters,
            filter_by_resource=filter_by_resource,
            filter_parameter=filter_parameter,
            sort_fields=sort_fields,
            resource_name=resource_type,
            resource_id=id_,
            server_url=url,
            auth_server_url=auth_server_url,
            auth_client_id=auth_client_id,
            auth_client_secret=auth_client_secret,
            auth_login_token=auth_login_token,
            auth_access_token=token_,
            auth_scopes=auth_scopes,
            include_only_properties=include_only_properties,
            separate_bundle_resources=separate_bundle_resources,
            expand_fhir_bundle=expand_fhir_bundle,
            accept_type=accept_type,
            content_type=content_type,
            additional_request_headers=additional_request_headers,
            accept_encoding=accept_encoding,
            extra_context_to_return=(
                {slug_column: service_slug} if slug_column and service_slug else None
            ),
            retry_count=retry_count,
            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
            limit=limit,
            log_level=log_level,
            use_data_streaming=use_data_streaming,
            graph_json=graph_json,
            refresh_token_function=refresh_token_function,
        )

    @staticmethod
    async def send_fhir_request_async(
        *,
        logger: Logger,
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        resource_name: str,
        resource_id: Optional[Union[List[str], str]],
        server_url: str,
        include_only_properties: Optional[List[str]],
        log_level: Optional[str],
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
        sort_fields: Optional[List[SortField]] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_access_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        separate_bundle_resources: bool = False,
        expand_fhir_bundle: bool = True,
        accept_type: Optional[str] = None,
        content_type: Optional[str] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        accept_encoding: Optional[str] = None,
        extra_context_to_return: Optional[Dict[str, Any]] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        limit: Optional[int] = None,
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> FhirGetResponse:
        """
        Sends a fhir request to the fhir client sdk


        :param logger:
        :param action:
        :param action_payload:
        :param additional_parameters:
        :param filter_by_resource:
        :param filter_parameter:
        :param resource_name:
        :param resource_id:
        :param server_url:
        :param include_only_properties:
        :param page_number:
        :param page_size:
        :param last_updated_after:
        :param last_updated_before:
        :param sort_fields:
        :param auth_server_url:
        :param auth_client_id:
        :param auth_client_secret:
        :param auth_login_token:
        :param auth_access_token:
        :param auth_scopes:
        :param separate_bundle_resources: separate bundle resources into a dict where each resourceType is a field
                                            with an array of results
        :param expand_fhir_bundle: expands the fhir bundle to create a list of resources
        :param accept_type: (Optional) Accept header to use
        :param content_type: (Optional) Content-Type header to use
        :param additional_request_headers: (Optional) Additional request headers to add in FHIR request
        :param accept_encoding: (Optional) Accept-encoding header to use
        :param extra_context_to_return: a dict to return with every row (separate_bundle_resources is set)
                                        or with FhirGetResponse
        :param retry_count: how many times to retry
        :param exclude_status_codes_from_retry: do not retry for these status codes
        :param limit: limit results to this
        :param log_level:
        :param use_data_streaming:
        :param graph_json:
        :param refresh_token_function: function to refresh token
        :return:
        :rtype:
        """

        fhir_client: FhirClient = get_fhir_client(
            logger=logger,
            server_url=server_url,
            auth_server_url=auth_server_url,
            auth_client_id=auth_client_id,
            auth_client_secret=auth_client_secret,
            auth_login_token=auth_login_token,
            auth_access_token=auth_access_token,
            auth_scopes=auth_scopes,
            log_level=log_level,
        )
        if not graph_json:  # graph_json passes this in the simulate_graph_async()
            fhir_client = fhir_client.separate_bundle_resources(
                separate_bundle_resources
            )
        fhir_client = fhir_client.expand_fhir_bundle(expand_fhir_bundle)
        if accept_type is not None:
            fhir_client = fhir_client.accept(accept_type)
        if content_type is not None:
            fhir_client = fhir_client.content_type(content_type)
        if accept_encoding is not None:
            fhir_client = fhir_client.accept_encoding(accept_encoding)
        if additional_request_headers is not None:
            logger.info(
                f"Additional Request Headers to be sent - {additional_request_headers}"
            )
            fhir_client = fhir_client.additional_request_headers(
                additional_request_headers
            )

        fhir_client = fhir_client.resource(resource_name)
        if resource_id:
            fhir_client = fhir_client.id_(resource_id)
            if filter_by_resource:
                fhir_client = fhir_client.filter_by_resource(filter_by_resource)
                if filter_parameter:
                    fhir_client = fhir_client.filter_parameter(filter_parameter)
        # add action to url
        if action:
            fhir_client = fhir_client.action(action)
            if action_payload:
                fhir_client = fhir_client.action_payload(action_payload)
        # add a query for just desired properties
        if include_only_properties:
            fhir_client = fhir_client.include_only_properties(
                include_only_properties=include_only_properties
            )
        if page_size and page_number is not None:
            fhir_client = fhir_client.page_size(page_size).page_number(page_number)
        if sort_fields is not None:
            fhir_client = fhir_client.sort_fields(sort_fields)

        if additional_parameters:
            fhir_client = fhir_client.additional_parameters(additional_parameters)

        # have to done here since this arg can be used twice
        if last_updated_before:
            fhir_client = fhir_client.last_updated_before(last_updated_before)
        if last_updated_after:
            fhir_client = fhir_client.last_updated_after(last_updated_after)

        if extra_context_to_return:
            fhir_client = fhir_client.extra_context_to_return(extra_context_to_return)

        if retry_count is not None:
            fhir_client = fhir_client.retry_count(retry_count)

        if exclude_status_codes_from_retry:
            fhir_client = fhir_client.exclude_status_codes_from_retry(
                exclude_status_codes_from_retry
            )

        if limit is not None:
            fhir_client = fhir_client.limit(limit=limit)

        if use_data_streaming:
            fhir_client = fhir_client.use_data_streaming(use_data_streaming)

        if refresh_token_function:
            fhir_client = fhir_client.refresh_token_function(refresh_token_function)

        return (
            await fhir_client.simulate_graph_async(
                id_=(
                    cast(str, resource_id)
                    if not isinstance(resource_id, list)
                    else resource_id[0]
                ),
                graph_json=graph_json,
                contained=False,
                separate_bundle_resources=separate_bundle_resources,
            )
            if graph_json
            else await fhir_client.get_async()
        )

    @staticmethod
    def json_str_to_list_str(json_str: str) -> List[str]:
        """
        at some point helix.fhir.client.sdk changed, and now it sends json string instead of list of json strings
        the PR: https://github.com/icanbwell/helix.fhir.client.sdk/pull/5
        this function converts the new returning format to old one
        """
        full_json = json.loads(json_str) if json_str else []
        if isinstance(full_json, list):
            return [json.dumps(item) for item in full_json]
        else:
            return [json_str]

    @staticmethod
    def get_batch_result(
        *,
        page_size: Optional[int],
        limit: Optional[int],
        server_url: Optional[str],
        action: Optional[str],
        action_payload: Optional[Dict[str, Any]],
        additional_parameters: Optional[List[str]],
        filter_by_resource: Optional[str],
        filter_parameter: Optional[str],
        resource_name: str,
        include_only_properties: Optional[List[str]],
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        sort_fields: Optional[List[SortField]],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_access_token: Optional[str],
        auth_scopes: Optional[List[str]],
        separate_bundle_resources: bool,
        expand_fhir_bundle: bool,
        accept_type: Optional[str],
        content_type: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        accept_encoding: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        log_level: Optional[str],
        error_view: Optional[str],
        ignore_status_codes: List[int],
        use_data_streaming: Optional[bool],
        graph_json: Optional[Dict[str, Any]],
        refresh_token_function: Optional[RefreshTokenFunction],
    ) -> GetBatchResult:
        resources: List[str] = []
        errors: List[str] = []
        if not page_size:
            page_size = limit
        # if paging is requested then iterate through the pages until the response is empty
        page_number: int = 0
        server_page_number: int = 0
        has_next_page: bool = True
        loop_number: int = 0
        assert server_url
        result: FhirGetResponse
        if use_data_streaming:
            result = AsyncHelper.run(
                FhirReceiverHelpers.send_fhir_request_async(
                    logger=get_logger(__name__),
                    action=action,
                    action_payload=action_payload,
                    additional_parameters=additional_parameters,
                    filter_by_resource=filter_by_resource,
                    filter_parameter=filter_parameter,
                    resource_name=resource_name,
                    resource_id=None,
                    server_url=server_url,
                    include_only_properties=include_only_properties,
                    limit=limit,
                    last_updated_after=last_updated_after,
                    last_updated_before=last_updated_before,
                    sort_fields=sort_fields,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_access_token=auth_access_token,
                    auth_scopes=auth_scopes,
                    separate_bundle_resources=separate_bundle_resources,
                    expand_fhir_bundle=expand_fhir_bundle,
                    accept_type=accept_type,
                    content_type=content_type,
                    additional_request_headers=additional_request_headers,
                    accept_encoding=accept_encoding,
                    retry_count=retry_count,
                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                    log_level=log_level,
                    use_data_streaming=use_data_streaming,
                    graph_json=graph_json,
                    refresh_token_function=refresh_token_function,
                )
            )
            try:
                resources = FhirReceiverHelpers.json_str_to_list_str(
                    result.get_response_text()
                )
            except JSONDecodeError as e:
                if error_view:
                    errors.append(
                        json.dumps(
                            {
                                "url": result.url,
                                "status_code": result.status,
                                "error_text": str(e)
                                + " : "
                                + result.get_response_text(),
                            },
                            default=str,
                        )
                    )
                else:
                    raise FhirParserException(
                        url=result.url,
                        message="Parsing result as json failed",
                        json_data=result.get_response_text(),
                        response_status_code=result.status,
                        request_id=result.request_id,
                    ) from e
        else:
            while has_next_page:
                loop_number += 1
                result = AsyncHelper.run(
                    FhirReceiverHelpers.send_fhir_request_async(
                        logger=get_logger(__name__),
                        action=action,
                        action_payload=action_payload,
                        additional_parameters=additional_parameters,
                        filter_by_resource=filter_by_resource,
                        filter_parameter=filter_parameter,
                        resource_name=resource_name,
                        resource_id=None,
                        server_url=server_url,
                        include_only_properties=include_only_properties,
                        page_number=server_page_number,  # since we're setting id:above we can leave this as 0
                        page_size=page_size,
                        limit=limit,
                        last_updated_after=last_updated_after,
                        last_updated_before=last_updated_before,
                        sort_fields=sort_fields,
                        auth_server_url=auth_server_url,
                        auth_client_id=auth_client_id,
                        auth_client_secret=auth_client_secret,
                        auth_login_token=auth_login_token,
                        auth_access_token=auth_access_token,
                        auth_scopes=auth_scopes,
                        separate_bundle_resources=separate_bundle_resources,
                        expand_fhir_bundle=True,  # always expand bundles so we can aggregate resources properly.
                        # we'll convert back into a bundle at the end if necessary
                        accept_type=accept_type,
                        content_type=content_type,
                        additional_request_headers=additional_request_headers,
                        accept_encoding=accept_encoding,
                        retry_count=retry_count,
                        exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                        log_level=log_level,
                        use_data_streaming=use_data_streaming,
                        graph_json=graph_json,
                        refresh_token_function=refresh_token_function,
                    )
                )

                auth_access_token = result.access_token
                resources1: FhirResourceList = result.get_resources()
                if len(resources1) > 0:
                    # get id of last resource
                    json_resources: FhirResourceList = resources1
                    if result.status == 200:  # normal response
                        if len(json_resources) > 0:  # received any resources back
                            last_json_resource = json_resources[-1]
                            if result.next_url:
                                # if server has sent back a next url then use that
                                next_url: Optional[str] = result.next_url
                                next_uri: furl = furl(next_url)
                                additional_parameters = [
                                    f"{k}={v}" for k, v in next_uri.args.allitems()
                                ]
                                # remove any entry for id:above
                                additional_parameters = list(
                                    filter(
                                        lambda x: not x.startswith("_count")
                                        and not x.startswith("_element"),
                                        additional_parameters,
                                    )
                                )
                            elif "id" in last_json_resource:
                                # use id:above to optimize the next query
                                id_of_last_resource = last_json_resource["id"]
                                if not additional_parameters:
                                    additional_parameters = []
                                # remove any entry for id:above
                                additional_parameters = list(
                                    filter(
                                        lambda x: not x.startswith("id:above"),
                                        additional_parameters,
                                    )
                                )
                                additional_parameters.append(
                                    f"id:above={id_of_last_resource}"
                                )
                            else:
                                server_page_number += 1
                            resources = resources + [
                                json.dumps(r.dict(), cls=FhirJSONEncoder)
                                for r in resources1
                            ]
                        page_number += 1
                        if limit and 0 < limit <= len(resources):
                            has_next_page = False
                    else:
                        # Received an error
                        if result.status == 404 and loop_number > 1:
                            # 404 (not found) is fine since it just means we ran out of data while paging
                            pass
                        elif result.status not in ignore_status_codes:
                            raise FhirReceiverException(
                                url=result.url,
                                json_data=result.get_response_text(),
                                response_text=result.get_response_text(),
                                response_status_code=result.status,
                                message="Error from FHIR server",
                                request_id=result.request_id,
                            )
                        has_next_page = False
                else:
                    has_next_page = False

        return GetBatchResult(resources=resources, errors=errors)



from typing import Dict, Any, Callable

import fastjsonschema


class SchemaValidator:
    def __init__(self, schema: Dict[str, Any]) -> None:
        self.schema_validator: Callable[[Any], None] = fastjsonschema.compile(schema)

    def validate(self, content: Any) -> bool:
        """
        Checks the content against schema and throws error if validation fails
        :param content:
        """
        self.schema_validator(content)
        return True



import logging
from datetime import datetime
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

import pytest
from typing import List, Dict, Any

from aioresponses import aioresponses
from pyspark.sql.types import Row
from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor_spark import (
    FhirReceiverProcessorSpark,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_receiver_exception import (
    FhirReceiverException,
)


@pytest.mark.parametrize("use_data_streaming", [True, False])
async def test_get_all_resources_async(
    spark_session: SparkSession, use_data_streaming: bool
) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    # Create a sample DataFrame
    data = [("1", "abc"), ("2", "def")]
    df: DataFrame = spark_session.createDataFrame(data, ["resource_id", "access_token"])

    # Define parameters
    parameters = get_fhir_receiver_parameters()
    parameters.use_data_streaming = use_data_streaming

    with aioresponses() as m:
        # Mock the FHIR server responses
        if use_data_streaming:
            m.get(
                "http://fhir-server/Patient",
                body="""{"resource_type": "Patient", "id": "1"}\n{"resource_type": "Patient", "id": "2"}""",
            )
            m.get(
                "http://fhir-server/Patient?id:above=2",
                body="",
            )
        else:
            m.get(
                "http://fhir-server/Patient?_count=5&_getpagesoffset=0",
                payload={
                    "resourceType": "Bundle",
                    "entry": [
                        {"resource": {"id": "1", "resourceType": "Patient"}},
                    ],
                },
            )
            m.get(
                "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=1",
                payload={
                    "resourceType": "Bundle",
                    "entry": [{"resource": {"id": "2", "resourceType": "Patient"}}],
                },
            )
            m.get(
                "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=2",
                status=404,
            )

        # Call the method
        result: DataFrame = await FhirReceiverProcessorSpark.get_all_resources_async(
            df=df,
            parameters=parameters,
            delta_lake_table=None,
            last_updated_after=None,
            last_updated_before=None,
            batch_size=10,
            mode="overwrite",
            file_path=temp_folder,
            page_size=5,
            limit=None,
            progress_logger=None,
            view=None,
            error_view=None,
            logger=logging.getLogger(__name__),
        )

        result.show(truncate=False)
        # Collect the result
        result_data: List[Row] = result.collect()

        # Assert the results
        assert len(result_data) == 2
        # assert result_data[0]["resourceType"] == "Patient"
        assert result_data[0]["resource_id"] == "1"
        # assert result_data[1]["resourceType"] == "Patient"
        assert result_data[1]["resource_id"] == "2"


async def test_get_all_resources_not_found_non_streaming_async(
    spark_session: SparkSession,
) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    # Create a sample DataFrame
    data = [("1", "abc"), ("2", "def")]
    df: DataFrame = spark_session.createDataFrame(data, ["resource_id", "access_token"])

    # Define parameters
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server responses
        m.get("http://fhir-server/Patient?_count=5&_getpagesoffset=0", status=404)

        with pytest.raises(FhirReceiverException):
            # Call the method
            result: DataFrame = (
                await FhirReceiverProcessorSpark.get_all_resources_non_streaming_async(
                    df=df,
                    parameters=parameters,
                    last_updated_after=None,
                    last_updated_before=None,
                    mode="overwrite",
                    file_path=temp_folder,
                    page_size=5,
                    limit=None,
                    file_format="json",
                )
            )

            result.show(truncate=False)
            # Collect the result
            result_data: List[Row] = result.collect()

            # Assert the results
            assert len(result_data) == 2
            # assert result_data[0]["resourceType"] == "Patient"
            assert result_data[0]["resource_id"] == "1"
            # assert result_data[1]["resourceType"] == "Patient"
            assert result_data[1]["resource_id"] == "2"


async def test_get_all_resources_not_found_streaming_async(
    spark_session: SparkSession,
) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    # Create a sample DataFrame
    data = [("1", "abc"), ("2", "def")]
    df: DataFrame = spark_session.createDataFrame(data, ["resource_id", "access_token"])

    # Define parameters
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server responses
        m.get(
            "http://fhir-server/Patient",
            status=404,
        )

        # Call the method
        result: DataFrame = (
            await FhirReceiverProcessorSpark.get_all_resources_streaming_async(
                df=df,
                parameters=parameters,
                last_updated_after=None,
                last_updated_before=None,
                mode="overwrite",
                file_path=temp_folder,
                file_format="json",
                batch_size=10,
            )
        )

        result.show(truncate=False)
        # Collect the result
        result_data: List[Row] = result.collect()

        # Assert the results
        assert len(result_data) == 0


@pytest.mark.parametrize("use_data_streaming", [True, False])
async def test_get_all_resources_empty_bundle__async(
    spark_session: SparkSession, use_data_streaming: bool
) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    # Create a sample DataFrame
    data = [("1", "abc"), ("2", "def")]
    df: DataFrame = spark_session.createDataFrame(data, ["resource_id", "access_token"])

    # Define parameters
    parameters = get_fhir_receiver_parameters()
    parameters.use_data_streaming = use_data_streaming

    with aioresponses() as m:
        # Mock the FHIR server responses
        if use_data_streaming:
            m.get(
                "http://fhir-server/Patient",
                payload={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "timestamp": "2022-08-08T07:50:38.3838Z",
                    "total": 0,
                    "entry": [],
                },
            )
        else:
            m.get(
                "http://fhir-server/Patient?_count=5&_getpagesoffset=0",
                payload={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "timestamp": "2022-08-08T07:50:38.3838Z",
                    "total": 0,
                    "entry": [],
                },
            )

        # Call the method
        result: DataFrame = await FhirReceiverProcessorSpark.get_all_resources_async(
            df=df,
            parameters=parameters,
            delta_lake_table=None,
            last_updated_after=None,
            last_updated_before=None,
            batch_size=10,
            mode="overwrite",
            file_path=temp_folder,
            page_size=5,
            limit=None,
            progress_logger=None,
            view=None,
            error_view=None,
            logger=logging.getLogger(__name__),
        )

        result.show(truncate=False)
        # Collect the result
        result_data: List[Row] = result.collect()

        # Assert the results
        assert len(result_data) == 2
        # assert result_data[0]["resourceType"] == "Patient"
        assert result_data[0]["resource_id"] == "1"
        # assert result_data[1]["resourceType"] == "Patient"
        assert result_data[1]["resource_id"] == "2"


async def test_process_partition_single_row() -> None:
    parameters = get_fhir_receiver_parameters()

    input_values: List[Dict[str, Any]] = [{"id": "1", "token": "abc"}]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for result in FhirReceiverProcessorSpark.process_chunk(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=range(1),
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=parameters,
            additional_parameters=None,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']


async def test_process_partition_multiple_rows_bundle() -> None:
    parameters = get_fhir_receiver_parameters()

    input_values: List[Dict[str, Any]] = [{"id": "1", "token": "abc"}]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={
                "resourceType": "Bundle",
                "type": "searchset",
                "entry": [
                    {"resource": {"resourceType": "Patient", "id": "1"}},
                    {"resource": {"resourceType": "Patient", "id": "2"}},
                ],
            },
        )

        async for result in FhirReceiverProcessorSpark.process_chunk(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=range(1),
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=parameters,
            additional_parameters=None,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == [
                '{"resourceType": "Patient", "id": "1"}',
                '{"resourceType": "Patient", "id": "2"}',
            ]



class VendorResponseKeyError(Exception):
    """
    shows that something is wrong with the vendor and we should not continue calling it
    """

    def __init__(
        self,
        message: str = "Too many errors calling Address Standardization api,"
        " please check prior errors in the log for more details",
    ) -> None:
        self.message = message
        super().__init__(self.message)



import os
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fix_fhir_bundle_transformer.v1.fix_fhir_bundle_transformer import (
    FixFhirBundleTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_fix_fhir_bundle_transformer(spark_session: SparkSession) -> None:
    data_dir: Path = Path(__file__).parent.joinpath("./")
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    input_path = data_dir.joinpath("input").__str__()
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    output_path = data_dir.joinpath("temp").__str__()
    with ProgressLogger() as progress_logger:
        FixFhirBundleTransformer(
            input_path=input_path,
            output_path=output_path,
            resources_to_extract=["Patient", "Encounter", "Observation"],
            progress_logger=progress_logger,
        ).transform(df)
    # now read Patient
    resources_df = spark_session.read.json(os.path.join(output_path, "Patient"))
    resources_df.printSchema()
    assert resources_df.count() == 2
    assert (
        resources_df.orderBy("id").select("birthDate").collect()[0][0] == "1970-01-01"
    )



import json
import os
import random
from copy import deepcopy
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Optional
from unittest import mock
from unittest.mock import MagicMock

import boto3

import pytest
import requests
from moto import mock_aws
from pymongo.collection import Collection
from requests import Session

import pymongo


from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.document_db_cache_handler import (
    DocumentDBCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.mock_cache_handler import (
    MockCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardize_address import (
    StandardizeAddr,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response_key_error import (
    VendorResponseKeyError,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendors.melissa_standardizing_vendor import (
    MelissaStandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendors.mock_standardizing_vendor import (
    MockStandardizingVendor,
)

mongo_address = r"mongodb://mongo:27017/"
mongo_collection_name = "helix_address_cache"

raw_addr_obj = RawAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    zipcode="78753",
)

raw_addr_obj1 = RawAddress(
    address_id="100",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    zipcode="78753",
)

raw_addr_obj2 = RawAddress(
    address_id="11", line1="1137 Huntington Drive Ste B1, South Pasadena, CA 91030"
)

std_addr_obj = StandardizedAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    country="US",
    zipcode="78753",
    latitude="30.373400",
    longitude="-97.680000",
    formatted_address="8300 North Lamar Boulevard;Austin, TX 78753-5976",
)

std_addr_obj1 = StandardizedAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    country="US",
    zipcode="78753",
    latitude="30.373400",
    longitude="-97.680000",
    formatted_address="8300 North Lamar Boulevard;Austin, TX 78753-5976",
)

std_addr_obj2 = StandardizedAddress(
    address_id="11",
    line1="1137 Huntington Drive Ste B1",
    city="Los Angeles",
    state="CA",
    country="US",
    zipcode="91030-4582",
    latitude="34.100066",
    longitude="-118.155034",
    formatted_address="1137 Huntington Dr Ste B1;South Pasadena, CA 91030-4582",
)
response_data: Dict[str, Any] = {
    "Version": "3.0.1.160",
    "TransmissionReference": "mock test",
    "TransmissionResults": "",
    "TotalRecords": "2",
    "Records": [
        {
            "RecordID": "11",
            "Results": "AC11,AC16,AV25,GS06",
            "FormattedAddress": "1137 Huntington Dr Ste B1;South Pasadena, CA 91030-4582",
            "Organization": "",
            "AddressLine1": "1137 Huntington Dr Ste B1",
            "AddressLine2": "South Pasadena, CA 91030-4582",
            "AddressLine3": "",
            "AddressLine4": "",
            "AddressLine5": "",
            "AddressLine6": "",
            "AddressLine7": "",
            "AddressLine8": "",
            "SubPremises": "Ste B1",
            "DoubleDependentLocality": "",
            "DependentLocality": "",
            "Locality": "South Pasadena",
            "SubAdministrativeArea": "Los Angeles",
            "AdministrativeArea": "CA",
            "PostalCode": "91030-4582",
            "PostalCodeType": "",
            "AddressType": "H",
            "AddressKey": "91030458221",
            "SubNationalArea": "",
            "CountryName": "United States of America",
            "CountryISO3166_1_Alpha2": "US",
            "CountryISO3166_1_Alpha3": "USA",
            "CountryISO3166_1_Numeric": "840",
            "CountrySubdivisionCode": "US-CA",
            "Thoroughfare": "Huntington Drive",
            "ThoroughfarePreDirection": "",
            "ThoroughfareLeadingType": "",
            "ThoroughfareName": "Huntington",
            "ThoroughfareTrailingType": "Drive",
            "ThoroughfarePostDirection": "",
            "DependentThoroughfare": "",
            "DependentThoroughfarePreDirection": "",
            "DependentThoroughfareLeadingType": "",
            "DependentThoroughfareName": "",
            "DependentThoroughfareTrailingType": "",
            "DependentThoroughfarePostDirection": "",
            "Building": "",
            "PremisesType": "",
            "PremisesNumber": "1137",
            "SubPremisesType": "Ste",
            "SubPremisesNumber": "B1",
            "PostBox": "",
            "Latitude": "34.100066",
            "Longitude": "-118.155034",
            "DeliveryIndicator": "B",
            "MelissaAddressKey": "7245851331",
            "MelissaAddressKeyBase": "7381043882",
            "PostOfficeLocation": "",
            "SubPremiseLevel": "",
            "SubPremiseLevelType": "",
            "SubPremiseLevelNumber": "",
            "SubBuilding": "",
            "SubBuildingType": "",
            "SubBuildingNumber": "",
            "UTC": "UTC-08:00",
            "DST": "Y",
            "DeliveryPointSuffix": "",
            "CensusKey": "060374807042021",
            "Extras": "",
        }
    ],
}


def test_raw_address_to_export() -> None:
    addr_dict = raw_addr_obj.to_dict()
    addr_str = raw_addr_obj.to_str()
    addr_hash = raw_addr_obj.to_hash()
    assert addr_dict == {
        "address_id": "10",
        "country": "US",
        "zipcode": "78753",
        "state": "TX",
        "city": "Austin",
        "line2": "",
        "line1": "8300 N Lamar Blvd",
    }
    assert addr_str == "8300 N Lamar Blvd, Austin TX 78753 US"
    assert addr_hash == "aa866e8b20823e573dba7c369b778196d7e75208"


def test_std_address_to_export() -> None:
    addr_dict = std_addr_obj.to_dict()
    addr_str = std_addr_obj.address.formatted_address
    assert addr_dict == {
        "address_id": "10",
        "country": "US",
        "zipcode": "78753",
        "state": "TX",
        "city": "Austin",
        "line2": "",
        "line1": "8300 N Lamar Blvd",
        "county": "",
        "latitude": "30.373400",
        "longitude": "-97.680000",
        "formatted_address": "8300 North Lamar Boulevard;Austin, TX 78753-5976",
        "standardize_vendor": "melissa",
    }
    assert addr_str == "8300 North Lamar Boulevard;Austin, TX 78753-5976"


def read_cache_file() -> List[Dict[str, Any]]:
    test_file_path = Path(__file__).parent.joinpath("cache_test_data.json")
    with open(test_file_path) as f:
        r: List[Dict[str, Any]] = json.load(f)
        return r


@mock.patch.object(Session, "post")
def test_address_api_call(mocked_session: MagicMock) -> None:
    with mock_aws():
        # arrange
        os.environ["AWS_ACCESS_KEY_ID"] = "testing"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
        os.environ["AWS_SECURITY_TOKEN"] = "testing"
        os.environ["AWS_SESSION_TOKEN"] = "testing"
        ssm = boto3.client("ssm", region_name="us-east-1")
        ssm.put_parameter(
            Name="/prod/helix/external/melissa/license_key_credit",
            Value="fake license",
            Type="SecureString",
        )
        raw_addr_obj_copy = deepcopy(raw_addr_obj)
        raw_addr_obj_copy.set_id("100")
        raw_addrs = [raw_addr_obj_copy, raw_addr_obj1, raw_addr_obj2]
        expected_result = [std_addr_obj, std_addr_obj1, std_addr_obj2]
        arrange_mongo()
        response = requests.Response()
        response.status_code = requests.codes.ok
        response.raw = BytesIO(json.dumps(response_data).encode())
        mocked_session.return_value = response
        # act
        r: List[StandardizedAddress] = StandardizeAddr().standardize(
            raw_addrs,
            vendor_obj=MockStandardizingVendor(),
            cache_handler_obj=DocumentDBCacheHandler(),
        )
        assert 3 == len(r)
        # assert
        assert r[0].get_id() == "100"
        assert r[1].get_id() == "100"
        assert r[2].get_id() == "11"
        assert (
            r[0].address.formatted_address
            == expected_result[0].address.formatted_address
        )
        assert (
            r[1].address.formatted_address
            == expected_result[1].address.formatted_address
        )


def test_address_custom_api_call(mocked_post: Optional[Any] = None) -> None:
    raw_addrs = [raw_addr_obj2]
    expected_result = [std_addr_obj2]
    arrange_mongo()

    # act
    r: List[StandardizedAddress] = StandardizeAddr().standardize(
        raw_addrs,
        cache_handler_obj=MockCacheHandler(),
        vendor_obj=MelissaStandardizingVendor(
            license_key="mock", custom_api_call=lambda x: response_data
        ),
    )

    # assert
    assert r[0].get_id() == "11"
    assert (
        r[0].address.formatted_address == expected_result[0].address.formatted_address
    )


def test_documentdb_cache() -> None:
    # arrange
    arrange_mongo()
    raw_addrs = [raw_addr_obj, raw_addr_obj2]

    # act
    r = DocumentDBCacheHandler().check_cache(raw_addrs)
    print(r.found)
    print(r.not_found)
    assert len(r.found) == 1
    assert len(r.not_found) == 1
    assert r.found[0].get_id() == "10"
    assert r.not_found[0].get_id() == "11"


def arrange_mongo() -> None:
    col: Collection[Any] = (
        pymongo.MongoClient(mongo_address)
        .get_database(mongo_collection_name)
        .get_collection(mongo_collection_name)
    )
    col.delete_many({})
    file_contents: List[Dict[str, Any]] = read_cache_file()
    cached_vendor_response: Dict[str, Any] = list(file_contents)[0]
    col.insert_one(cached_vendor_response)


def test_standardize_stress_test() -> None:
    def mocked_api_call(addresses_to_lookup: List[RawAddress]) -> Dict[str, Any]:
        res = deepcopy(response_data)
        res["Records"] = [
            {
                "RecordID": a.address.address_id,
                "FormattedAddress": f"{a.to_str()}",
                "AddressLine1": a.address.line1,
                "AddressLine2": "",
                "Locality": a.address.city,
                "SubAdministrativeArea": "a county",
                "AdministrativeArea": a.address.state,
                "PostalCode": a.address.zipcode,
                "CountryISO3166_1_Alpha2": a.address.zipcode,
                "Latitude": "000",
                "Longitude": "-000",
            }
            for a in addresses_to_lookup
        ]
        return res

    # arrange
    # random addresses
    address_population_count = 1000
    sample_count = address_population_count // 10
    random_address = []
    for i in range(address_population_count):
        random_id = random.randint(1, sample_count)
        random_address.append(
            {
                "id": str(random_id),
                "line1": f"line1 of {random_id}",
                "city": f"city{random_id}",
                "zip": f"zipcode {random_id}",
                "state": f"state {random_id}",
                "country": "US",
                "lat": f"lat{random_id}",
                "long": f"long{random_id}",
            }
        )
    # create raw address objects
    raw_addr_objs = [
        RawAddress(
            address_id=a["id"],
            line1=a["line1"],
            city=a["city"],
            state=a["state"],
            country=a["country"],
            zipcode=a["zip"],
        )
        for a in random_address
    ]
    # creat expected std response - for future use
    # std_addr_objs = [
    #     StdAddress(
    #         address_id=a["id"],
    #         line1=a["line1"],
    #         city=a["city"],
    #         state=a["state"],
    #         country=a["country"],
    #         zipcode=a["zip"],
    #         latitude=a["lat"],
    #         longitude=a["long"],
    #         formatted_address=f'a["line1"] a["city"] a["country"] a["zip"]',
    #     )
    #     for a in random_address
    # ]
    pymongo.MongoClient(mongo_address).get_database(
        mongo_collection_name
    ).get_collection(mongo_collection_name).delete_many({})

    # act
    r = StandardizeAddr().standardize(
        raw_addresses=raw_addr_objs,
        vendor_obj=MelissaStandardizingVendor(
            license_key="mock", custom_api_call=mocked_api_call
        ),
        cache_handler_obj=DocumentDBCacheHandler(),
    )
    # assert
    assert len(r) == address_population_count
    assert (
        pymongo.MongoClient(mongo_address)
        .get_database(mongo_collection_name)
        .get_collection(mongo_collection_name)
        .count_documents({})
        == sample_count
    )


def test_bad_request() -> None:
    def mocked_api_call(addresses_to_lookup: List[RawAddress]) -> Dict[str, Any]:
        res = deepcopy(response_data)
        res["Records"] = [
            {
                "RecordID": a.address.address_id,
                "FormattedAddress": "",
                "AddressLine1": "",
                "AddressLine2": "",
                "Locality": "",
                "SubAdministrativeArea": "",
                "AdministrativeArea": "",
                "PostalCode": "",
                "CountryISO3166_1_Alpha2": "US",
                "Latitude": "",
                "Longitude": "",
            }
            for a in addresses_to_lookup
        ]
        return res

    # arrange
    # random addresses
    # create raw address objects
    empty_raw_address = RawAddress(
        address_id="1",
        line1="",
        city="",
        state="",
        country="US",
        zipcode="",
    )
    # creat expected std response

    pymongo.MongoClient(mongo_address).get_database(
        mongo_collection_name
    ).get_collection(mongo_collection_name).delete_many({})

    # act
    r = StandardizeAddr().standardize(
        raw_addresses=[empty_raw_address],
        vendor_obj=MelissaStandardizingVendor(
            license_key="mock", custom_api_call=mocked_api_call
        ),
        cache_handler_obj=DocumentDBCacheHandler(),
    )
    # assert
    assert r[0].get_id() == "1"


def test_none_cache() -> None:
    raw_addrs = [raw_addr_obj, raw_addr_obj2]
    r = MockCacheHandler().check_cache(raw_addrs)
    assert len(r.not_found) == 2
    assert r.not_found[0].get_id() == "10"


@mock.patch.object(Session, "post")
def test_vendor_http_error_call(mocked_session: MagicMock) -> None:
    response = requests.Response()
    response.status_code = 500
    response.raw = BytesIO(json.dumps({}).encode())
    mocked_session.return_value = response

    # arrange
    raw_addrs = [raw_addr_obj]

    # act / assert
    with pytest.raises(requests.exceptions.HTTPError):
        StandardizeAddr().standardize(
            raw_addrs,
            cache_handler_obj=MockCacheHandler(),
            vendor_obj=MelissaStandardizingVendor(license_key="mock"),
        )


@mock.patch.object(Session, "post")
def test_vendor_empty_response_call(mocked_session: MagicMock) -> None:
    response = requests.Response()
    response.status_code = 200
    response.raw = BytesIO(json.dumps({}).encode())
    mocked_session.return_value = response

    # arrange
    raw_addrs = [raw_addr_obj]
    vendor = MelissaStandardizingVendor(license_key="mock")
    vendor._RESPONSE_KEY_ERROR_THRESHOLD = 0
    # act / assert
    with pytest.raises(VendorResponseKeyError):
        StandardizeAddr().standardize(
            raw_addrs, cache_handler_obj=MockCacheHandler(), vendor_obj=vendor
        )



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from spark_fhir_schemas.r4.resources.patient import PatientSchema

from spark_pipeline_framework.transformers.fhir_reader.v1.fhir_reader import FhirReader
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_fhir_reader(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    test_files_dir: Path = data_dir.joinpath("test_files/patients")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # Act
    FhirReader(
        view="patients", schema=PatientSchema.get_schema(), file_path=test_files_dir
    ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("patients")
    result_df.printSchema()
    result_df.show(truncate=False)

    assert result_df.count() == 2



from typing import Any, Dict, Optional, Union

from pyspark.ml.param import Param
from pyspark.sql import DataFrameWriter
from pyspark.sql.dataframe import DataFrame
from pyspark.errors import AnalysisException
from pyspark.sql.streaming.readwriter import DataStreamWriter

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters


class FrameworkBaseExporter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = FileWriteModes.MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: Optional[int] = None,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ):
        assert mode in FileWriteModes.MODE_CHOICES

        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.stream: Param[bool] = Param(self, "stream", "")
        self._setDefault(stream=stream)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        name: Optional[str] = self.getName()
        format_: str = self.getFormat()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        stream: bool = self.getStream()
        limit: Optional[int] = self.getLimit()

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        format_ = "delta" if delta_lake_table else format_

        with ProgressLogMetric(
            name=f"{name or view}_{format_}_exporter", progress_logger=progress_logger
        ):
            try:
                writer: Union[DataFrameWriter, DataStreamWriter]
                if view:
                    df = df.sparkSession.table(view)
                if limit is not None and limit >= 0:
                    df = df.limit(limit)
                writer = df.write if not stream else df.writeStream

                writer = writer.format(format_)

                if isinstance(writer, DataFrameWriter):
                    mode = self.getMode()
                    writer = writer.mode(mode)

                for k, v in self.getOptions().items():
                    writer.option(k, v)

                if progress_logger:
                    progress_logger.log_params(self.getOptions())

                if isinstance(writer, DataStreamWriter):
                    writer.start()
                else:
                    writer.save()

            except AnalysisException as e:
                self.logger.error(f"Failed to write to {format_}")
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming
    def getFormat(self) -> str:
        raise NotImplementedError

    # noinspection PyPep8Naming
    def getOptions(self) -> Dict[str, Any]:
        raise NotImplementedError

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStream(self) -> bool:
        return self.getOrDefault(self.stream)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDeltaLakeTable(self) -> Optional[str]:
        return self.getOrDefault(self.delta_lake_table)



from typing import List, Type

import structlog


from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response import (
    VendorResponse,
)

logger = structlog.get_logger(__file__)


class MockCacheHandler(CacheHandler):
    def _get_vendor_class(self, vendor_name: str) -> Type[StandardizingVendor]:
        return StandardizingVendor

    def check_cache(self, raw_addresses: List[RawAddress]) -> CacheResult:
        return CacheResult(found=[], not_found=raw_addresses)

    def save_to_cache(self, vendor_responses: List[VendorResponse]) -> None:
        pass



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType, IntegerType, StringType

from create_spark_session import clean_spark_session
from spark_pipeline_framework.transformers.framework_fixed_width_loader.v1.framework_fixed_width_loader import (
    FrameworkFixedWidthLoader,
    ColumnSpec,
)


def test_can_load_fixed_width(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test.txt')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkFixedWidthLoader(
        view="my_view",
        file_path=test_file_path,
        columns=[
            ColumnSpec(column_name="id", start_pos=1, length=3, data_type=StringType()),
            ColumnSpec(
                column_name="some_date", start_pos=4, length=8, data_type=StringType()
            ),
            ColumnSpec(
                column_name="some_string",
                start_pos=12,
                length=3,
                data_type=StringType(),
            ),
            ColumnSpec(
                column_name="some_integer",
                start_pos=15,
                length=4,
                data_type=IntegerType(),
            ),
        ],
    ).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM my_view")

    result.show()

    # Assert
    assert result.count() == 2
    assert result.collect()[0][0] == "001"
    assert result.collect()[1][0] == "002"
    assert result.collect()[0][1] == "01292017"
    assert result.collect()[1][1] == "01302017"
    assert result.collect()[0][2] == "you"
    assert result.collect()[1][2] == "me"
    assert result.collect()[0][3] == 1234
    assert result.collect()[1][3] == 5678
